#include "CruCore.h"

#include <QElapsedTimer>
#include <QThread>

#include "CruFsm.h"
#include "CruSignalRouter.h"

#include "base/DeviceIndexEnabled.h"
#include "base/RunDocument.h"
#include "mldiscover/DeviceDispatcher.h"
#include "lib-common/ReadoutAppConfigDb.h"
#include "lib-common/ReadoutAppConfigDispatcher.h"

CruCore::CruCore(QObject *parent)
    : ReadoutAppCore(parent)
    , cruFsm(new CruFsm(this))
    , configDb(new ReadoutAppConfigDb(this))
{
    qRegisterMetaType<DeviceIndex>();
    qRegisterMetaType<LldpInfoCache>();
    qRegisterMetaType<RunDocument>();
    qRegisterMetaType<ModularDeviceStatus>();

    auto signalRouter = &CruSignalRouter::instance();
    signalRouter->setDeviceDispatcher(deviceDispatcher);
    signalRouter->setConfigDb(configDb);
    signalRouter->setCore(this);
    signalRouter->setReadoutAppConfigDispatcher(
                new ReadoutAppConfigDispatcher(this));
    signalRouter->setCruFsm(cruFsm);
    signalRouter->connectAll();

    initCruProgDescr();
}

const BaseFsm *CruCore::getFsm() const
{
    return cruFsm;
}

void CruCore::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    Q_UNUSED(config_key)
    if(config == newConfig)
        return;
    const auto &oldSet = config.getDeviceIndexEnabledMap().getEnabled();
    const auto &newSet = newConfig.getDeviceIndexEnabledMap().getEnabled();
    config = newConfig;

    const auto &toRemove = oldSet-newSet;
    if(!toRemove.isEmpty()) {
        for (const auto &i : toRemove)
            removeDevice(i);
    }

    const auto &toCreate = newSet-oldSet;
    for (const auto &i : toCreate) {
        if (! devices.contains(i))
            addDevice(i);
    }
    updateCruProgDescrOptions();
    emit allowRemoteControl(config.allowRemoteControl);
}

void CruCore::performLocalStart()
{
    if(cruFsm->getState() == FsmState::Ready) {
        rcRunId = {};
        rcRunId.uuid = QUuid::createUuid();
        runStartEvNum.clear();
        cruFsm->setRequestState(FsmState::Configure);
    } else {
        qWarning() << "Can't" << __func__ << "in state:" <<cruFsm->getState();
        cruFsm->setRequestState(FsmState::Error);
    }
}

void CruCore::performLocalStop()
{
    if(cruFsm->getState() == FsmState::Run) {
        cruFsm->setRequestState(FsmState::Stopping);
    } else {
        qWarning() << "Can't" << __func__ << "in state:" <<cruFsm->getState();
        cruFsm->setRequestState(FsmState::Error);
    }
}

void CruCore::remoteStartRequest(const RcRunId &newRunId, const QVariant &startEvNum)
{
    if(cruFsm->getState() != FsmState::Ready) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
            .arg(fsmStateToString(FsmState::Ready))
            .arg(fsmStateToString(cruFsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(newRunId, msg);
        return;
    }

    rcRunId = newRunId;
    runStartEvNum = startEvNum;
    cruFsm->setRequestState(FsmState::Configure);
}

void CruCore::remoteStopRequest(const RcRunId &newRunId, const QVariant &stopEvNum)
{
    Q_UNUSED(stopEvNum)
    if(rcRunId.runIndex != newRunId.runIndex) {
        qWarning()<< QString("Got stop request with wrong run index: runIndex (%1); "
                             "remote control runIndex (%2)")
                     .arg(newRunId.runIndex)
                     .arg(newRunId.runIndex);
    }

    if(cruFsm->getState() == FsmState::Run) {
        cruFsm->setRequestState(FsmState::Stopping);
    } else {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Run))
                              .arg(fsmStateToString(cruFsm->getState()));
        qWarning() << msg;
        emit sendRunWarning(newRunId, msg);
        cruFsm->setRequestState(FsmState::Error);
    }
}

void CruCore::onRemoteStatusUpdated(const RemoteControlServerStatus &status)
{
    bool changed = progDescr.updateProgramInterface(
                ProgramInterfaceRemoteControl, 0,
                status.tcpPort, status.listening, !status.portInUse,
                status.peerHost, status.peerPort);
    if(changed)
        emit updateProgramDescription(progDescr);

    if(status.portInUse) {
        QString ready4run_str = "Ready for run";
        for(DeviceIndex &index : devices.keys()){
            if(!dev_status.value(index).connected){
                ready4run_str = "Not ready for run";
            }
        }
    } else if(lastRcStatus.portInUse) {
        qWarning() << "Remote control disconnected.";

        if(isRun()) {
            qCritical() << "Stopping current run";
            cruFsm->setRequestState(cruFsm->getState() == FsmState::Run
                         ? FsmState::Stopping : FsmState::Error);
        }
    }
    lastRcStatus = status;
}

void CruCore::onFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    ReadoutAppCore::onFsmStateChange(state, oldState);

    switch (state) {
    case FsmState::Error :
        if(oldState != FsmState::Init) {
            runStartEvNum.clear();
            config.setRunState(false);
            emit configChanged(config);
        }
        break;
    case FsmState::Ready :
    {
        switch (oldState) {
        case FsmState::Stopping:
            qInfo() << QString("%1 completed successfully")
                       .arg(fsmStateToString(oldState));
            break;
        default:{}
        }
        break;
    }
    case FsmState::Configure :
    {
        cruFsm->setRunId(rcRunId);
        config.setRunState(false);
        emit configChanged(config);
        emit preRunReset(runStartEvNum, rcRunId);

        cruFsm->setRequestState(FsmState::Starting);
        break;
    }
    case FsmState::Starting :
    {
        emit clearStats();
        config.setRunState(true);
        emit configChanged(config);
        cruFsm->setRequestState(FsmState::Run);
        break;
    }
    case FsmState::Run :
    {
        qInfo() << QString("Start completed successfully");
        sendRunDocument();
        break;
    }
    case FsmState::Stopping :
    {
        runStartEvNum.clear();
        config.setRunState(false);
        emit configChanged(config);
        cruFsm->setRequestState(FsmState::Ready);
        break;
    }
    default:;
    }
}

void CruCore::addDevice(const DeviceIndex &index)
{
    try {
        devices.insert(index, QSharedPointer<CruDeviceGlobal>(new CruDeviceGlobal(index)));
    } catch (std::exception &e){
        qCritical() << QString("[%1] Error on connect to device: %2")
                       .arg(index.getDeviceIdName()).arg(e.what());
    }
}

void CruCore::removeDevice(const DeviceIndex &index)
{
    devices.remove(index);
}

void CruCore::sendRunDocument() const
{
    auto cfgObj = QVariant::fromValue(config).value<QJsonObject>();
    RunDocument rd(rcRunId, configDb->getClientIndex(), cfgObj);
    emit sendRunDocumentToRc(rd);
}

void CruCore::deviceDisconnected(const DeviceIndex &index)
{
    if(isRun()){
        sendRunError(rcRunId, index.getIdent() + " disconnected");
    }
}

void CruCore::deviceStatusUpdated(const ModularDeviceStatus &status)
{
    dev_status[status.getDeviceIndex()] = status;
    bool device_disconnected = false;
    for(DeviceIndex &index : dev_status.keys()){
        if(!dev_status[index].connected)
            device_disconnected = true;
    }
    if(isRun() && device_disconnected){
        sendRunError(rcRunId, status.getDeviceIdent() + " disconnected");
    }
}

void CruCore::initCruProgDescr()
{
    auto cs = configDb->getConfigSelector();
    progDescr.setClientIndex(cs.getClientIndex());

    //todo: transfer it from libs.pro
#ifdef GIT_DESCR
    progDesc.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    progDesc.ver_date = GIT_LAST_COMMIT;
#endif

    emit updateProgramDescription(progDescr);
}

void CruCore::updateCruProgDescrOptions()
{
    QStringList devList;
    for(const auto &dg : devices)
        devList << dg->getDeviceIndex().toString();
    progDescr.options["DevList"] = devList.join(";");
    progDescr.options["run"] = QVariant(isRun()).toString();

    emit updateProgramDescription(progDescr);
}
