#include "HrbModel.h"

#include <QColor>
#include <QHostAddress>
#include <QPalette>
#include <QDateTime>


#include "base/DeviceIndexEnabled.h"
#include "config/HrbAppConfig.h"
#include "device-modules/ModularDeviceConfig.h"
#include "config/HrbConfigDispatcher.h"
#include "hardware/ClockUtil.h"
#include "hardware/HrbStatus.h"
#include "mldiscover/LldpDiscovery.h"

#include "HrbModelColumns.h"
#include "HrbSignalRouter.h"
#include "StatisticDialog.h"

struct HrbTableEntry
{
    ModularDeviceConfig config;
    HrbStatus status;
    LldpInfo lldp;
};

class HrbMetaTable : public QMap<DeviceIndex, HrbTableEntry>
{
public:
    int rowOf(const DeviceIndex &deviceIndex) const;
    DeviceIndex indexOf(int row) const;
};

int HrbMetaTable::rowOf(const DeviceIndex &deviceIndex) const
{
    return std::distance(begin(), find(deviceIndex));
}

DeviceIndex HrbMetaTable::indexOf(int row) const
{
    int r = 0;
    for (auto i : keys())
        if (r++ == row)
            return i;
    return DeviceIndex();
}

HrbModel::HrbModel(QObject *parent)
    : QAbstractTableModel(parent)
    , devices(new HrbMetaTable())
    , appCfg(new HrbAppConfig())
{
    HrbSignalRouter::instance().setHrbModel(this);
    HrbConfigDispatcher::instance().subscribeConfig(this);
    HrbConfigDispatcher::instance().subscribeHrbModel(this);
}

HrbModel::~HrbModel()
{
    delete devices;
    delete appCfg;
}

void HrbModel::setEnabled(bool en)
{
    enabled = en;
    sendDataChanged();

}

void HrbModel::invalidateRow(const DeviceIndex &index)
{
    int row = devices->rowOf(index);
    if (row == -1)
        return;

    auto indexFrom = createIndex(row, 0);
    auto indexTo = createIndex(row, columnCount()-1);
    emit dataChanged(indexFrom, indexTo);
}

void HrbModel::sendDataChanged(QVector<int> roles)
{
    emit dataChanged(index(0,0), index(rowCount()-1, columnCount()-1), roles);
}

void HrbModel::lldpCacheUpdated(const LldpInfoCache &lldp)
{
    for(auto it=devices->begin(); it!=devices->end(); ++it) {
        const auto &devLldp = lldp[it.key()];
        if(it->lldp == devLldp)
            continue;
        it->lldp = devLldp;
        invalidateRow(it.key());
    }
}

void HrbModel::setDevConfig(const ModularDeviceConfig &config)
{
    if (config == devices->value(config.index).config)
        return;
    (*devices)[config.index].config = config;
    invalidateRow(config.index);
}

void HrbModel::deviceStatusUpdated(const HrbStatus &status)
{
    const auto &deviceIndex = status.getDeviceIndex();
    if (status == devices->value(deviceIndex).status)
        return;
    (*devices)[deviceIndex].status = status;
    invalidateRow(deviceIndex);
}

void HrbModel::addDevice(const ModularDeviceConfig &cfg)
{
    const auto &index = cfg.index;
    auto tmp = *devices;
    tmp[index];
    int r = tmp.rowOf(index);
    beginInsertRows(QModelIndex(), r, r);
    auto &tte = (*devices)[index];
    tte.config = cfg;
    endInsertRows();
}

void HrbModel::removeDevice(const DeviceIndex &index)
{
    int row = devices->rowOf(index);
    Q_ASSERT(row!=-1);
    if(row==-1)
        return;;
    beginRemoveRows(QModelIndex(), row, row);
    devices->remove(index);
    endRemoveRows();
}

void HrbModel::setConfig(const ConfigKey &configKey, const HrbAppConfig &cfg)
{
    Q_UNUSED(configKey);
    if(*appCfg == cfg)
        return;
    const auto &newSet = cfg.getDeviceIndexEnabledMap(DeviceFilter::HRB).getEnabled();
    const auto &oldSet = appCfg->getDeviceIndexEnabledMap(DeviceFilter::HRB).getEnabled();
    *appCfg = cfg;

    for(auto i : oldSet-newSet)
        removeDevice(i);

    for(auto i : newSet-oldSet)
        addDevice(cfg.hrb[i]);

    for(auto i : oldSet & newSet) {
        Q_ASSERT(devices->contains(i));
        auto &oldCfg = (*devices)[i].config;
        const auto &newCfg = cfg.hrb[i];
        if(oldCfg == newCfg)
            continue;
        oldCfg = newCfg;
        invalidateRow(i);
    }

    if(showAveragedBusy != cfg.showAveragedBusy) {
        showAveragedBusy = cfg.showAveragedBusy;
        if(rowCount()) {
            for(auto c: {COL_HRB_BUSY_TIME, COL_HRB_BUSY_MSTR_TIME} ) {
                auto indexFrom = createIndex(0, c);
                auto indexTo = createIndex(rowCount()-1, c);
                emit dataChanged(indexFrom, indexTo);
            }
        }
    }
}

bool HrbModel::isChCol(int col)
{
    return col >= COL_HRB_ASD_FIRST && col <= COL_HRB_ASD_LAST;
}

int HrbModel::rowCount(const QModelIndex &) const
{
    return  devices->size();
}

int HrbModel::columnCount(const QModelIndex &) const
{
    return COL_HRB_TOTAL;
}

Qt::ItemFlags HrbModel::flags(const QModelIndex & index) const
{
    auto f = QAbstractTableModel::flags(index);

    if(isChCol(index.column()))
        f |= Qt::ItemIsEditable;

    if(!enabled)
        f &= ~Qt::ItemIsEnabled;

    return f;
}

QVariant HrbModel::data(const QModelIndex &index, int role) const
{
    auto column = index.column();
    auto row = index.row();
    auto deviceIndex = devices->indexOf(row);
    if (deviceIndex.isNull())
        return QVariant();

    switch (role) {
    case Qt::BackgroundRole: return data_BackgroundRole(deviceIndex, column);
    case Qt::DisplayRole: return data_DisplayRole(deviceIndex, column);
    case Qt::EditRole: return data_EditRole(deviceIndex, column);
    }
    return QVariant();
}

QVariant HrbModel::data_BackgroundRole(const DeviceIndex &index, int) const
{
    const auto device = (*devices)[index];
    // const auto &config = device.config;
    const auto &status = device.status;

    if(!status.connected)
        return QColor(Qt::lightGray);
    return QVariant();
}

QVariant HrbModel::data_DisplayRole(const DeviceIndex &index, int column) const
{
    const auto device = (*devices)[index];

    if(isChCol(column)) {
        const auto &config = device.config;
        const auto ch = column-COL_HRB_ASD_FIRST;
        return config.hrb.thr[ch];
    }

    const auto &status = device.status;
    const auto &lldpInfo = device.lldp;

    switch (column) {
    case COL_HRB_ID:
        return index.getSerialStr();
    case COL_HRB_IP:
        return lldpInfo.common.ip_addr ? QHostAddress(lldpInfo.common.ip_addr).toString() : QVariant();
    case COL_HRB_TIME:
        return StatisticDialog::getDateTimeFromClock(status.systemTs).toString("hh:mm:ss");
    case COL_HRB_TEMP:
        return TempsInfo(status.getTemps()).brief();
    case COL_HRB_SPILL:
        return status.spillNum;
    case COL_HRB_EVENT:
        return status.evNum;
    case COL_HRB_GLOB_EVENT:
        return status.globEvNum;
    case COL_HRB_WRITE:
        return QVariant(); // todel?
    case COL_HRB_PASS:
        return status.statistic.trigPass;
    case COL_HRB_BLOCK:
        return status.statistic.trigBlock;
    case COL_HRB_IDLE_TIME:
        return getTimeStr(status.statistic.idleTimer*HRB_CLK_SEC*1e3);
    case COL_HRB_BUSY_TIME:
    {
        auto t_us = status.statistic.busyTimer*HRB_CLK_SEC*1e6;
        if(showAveragedBusy) {
            if(status.globEvNum)
                return QString("%1 us").arg(t_us/status.globEvNum, 0, 'g', 3);
            else
                return QString();
        } else
            return getTimeStr(t_us/1e3);
    }
    case COL_HRB_IDLE_MSTR_TIME:
        return getTimeStr(status.statistic.idleMStreamTimer*HRB_CLK_SEC*1e3);
    case COL_HRB_BUSY_MSTR_TIME:
    {
        quint64 t = status.statistic.busyMStreamTimer*HRB_CLK_SEC*1e6;
        if(showAveragedBusy){
            if(status.globEvNum)
                return QString("%1 us").arg(1.*t/status.globEvNum, 0, 'g', 3);
            else
                return QString();
        } else
            return getTimeStr(t/1e3);
    }
    case COL_HRB_OCCUPANCY:
    {
        const auto &sumT = status.statistic.idleTimer+status.statistic.busyTimer;
        if(!sumT)
            return QString();
        double occupancy = 100.*status.statistic.busyTimer/sumT;
        return QString("%1%").arg(occupancy,0,'f',2);
    }
    default:;
    }
    return QVariant();
}

QVariant HrbModel::data_EditRole(const DeviceIndex &index, int column) const
{
    if(!isChCol(column))
        return QVariant();

    const auto device = (*devices)[index];
    const auto &config = device.config;
    const auto ch = column-COL_HRB_ASD_FIRST;
    return config.hrb.thr[ch];
}

QVariant HrbModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (role != Qt::DisplayRole)
        return QVariant();
    if (orientation != Qt::Horizontal)
        return QVariant();

    if(isChCol(section)) {
        const auto ch = section-COL_HRB_ASD_FIRST;
        return QString("ASD %1").arg(ch);
    }

    switch (section) {
    case COL_HRB_ID: return tr("Id");
    case COL_HRB_IP: return tr("Ip");
    case COL_HRB_TIME: return tr("Time");
    case COL_HRB_TEMP: return tr("Temp");
    case COL_HRB_SPILL: return tr("Spill");
    case COL_HRB_EVENT: return tr("Event");
    case COL_HRB_GLOB_EVENT: return tr("GlobEvent");
    case COL_HRB_WRITE: return tr("Write");
    case COL_HRB_PASS: return tr("Pass");
    case COL_HRB_BLOCK: return tr("Block");
    case COL_HRB_IDLE_TIME: return tr("Idle time");
    case COL_HRB_BUSY_TIME: return tr("Busy time");
    case COL_HRB_IDLE_MSTR_TIME: return tr("Idle time MStr");
    case COL_HRB_BUSY_MSTR_TIME: return tr("Busy time MStr");
    case COL_HRB_OCCUPANCY: return tr("Occupancy");
    }
    return QVariant();
}

bool HrbModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    auto column = index.column();
    auto row = index.row();
    auto deviceIndex = devices->indexOf(row);
    if (deviceIndex.isNull())
        return false;

    auto &device = (*devices)[deviceIndex];
    auto &config = device.config;

    if (role != Qt::EditRole)
        return false;

    if(!isChCol(column))
        return false;
    const auto &ch = column-COL_HRB_ASD_FIRST;
    config.hrb.thr[ch] = value.toDouble();

    emit dataChanged(index, index);
    emit hrbDeviceConfigChanged(config);
    return true;
}
