#include "Analizer.h"
#include "RingBuffer.h"
#include "lib-common/MpdRawTypes.h"

#include <QDebug>
#include <QElapsedTimer>
#include <QTimer>

namespace  {
const int LOOP_SIZE = 1000;
constexpr const int MPD_HDR_SIZE = sizeof(MpdHdr);
constexpr const int MPD_OLD_HDR_SIZE = sizeof(MpdEventHdr);
}

Analizer::Analizer(ClientIndex _index, RingBuffer<char> *dataRB, RingBuffer<EventMetaInfo> *evRB):
    AnalyzerFsm(nullptr),
    analizedSize(0),
    byteRB(dataRB),
    eventRB(evRB),
    lastMetaInfo(nullptr),
//    newRawDataWC(_newDataWC),
    enabled(false),
    index(_index),
    runElapsedTimer(new QElapsedTimer()),
    runTimer(new QTimer(this)),
    statTimer(new QTimer(this)),
    stat(index)
{
    setObjectName(__func__+(":"+index.toString()));

    connect(this, &BaseFsm::stateChanged, this, &Analizer::onFsmStateChanged);

    connect(runTimer, &QTimer::timeout, this, &Analizer::run);
    runTimer->setInterval(0);
    runTimer->setSingleShot(true);
    connect(statTimer, &QTimer::timeout, this, &Analizer::sentStat);
    statTimer->start(500);
}

Analizer::~Analizer()
{
    delete runElapsedTimer;
}

void Analizer::onFsmStateChanged(const FsmState &state, const FsmState &prev)
{
    switch (state) {
    case FsmState::Error :
    {
        if(prev==FsmState::Run || prev==FsmState::Configure || enabled) {
            doStateAction(FsmState::Ready);
        }
        break;
    }
    case FsmState::Configure:
    {
        qDebug() << QString("Analizer for #%1 was resetted")
                    .arg(index.toString());
        break;
    }
    default:;
    }
}

void Analizer::run()
{
    bool badSync=false;
    runElapsedTimer->start();
    while(enabled && timeLeftForRun()>0){
        if(!cleanUpFreedSpace())
            continue; // There is no free space in eventRB
        Q_ASSERT(eventRB->waitFreeSpace(1,0)); // Just to enshure that there is enough free space
        EventMetaInfo *i = eventRB->getWritePointer();
        Q_ASSERT(i->free);

        if(analizedSize+MPD_OLD_HDR_SIZE<0) {
            eventRB->waitTrashSpace(1, timeLeftForRun());
            continue;
        }
        // Ready to read new raw data
        if(byteRB->getDataSize() < analizedSize+MPD_OLD_HDR_SIZE) {
            if(!byteRB->waitDataSpace(analizedSize+MPD_OLD_HDR_SIZE, timeLeftForRun()))
                break;
        }

        i->reset();
        byteRB->readData(reinterpret_cast<char*>(&mpdHdr),
                         analizedSize, MPD_HDR_SIZE);
        switch (mpdHdr.sync) {
        case MPD_SYNC_NORM:
        case MPD_SYNC_NORM_TLV:
//            stat.lastEvNum = mpdEvHdr.evNum;
            byteRB->readData(reinterpret_cast<char*>(&i->evNum),
                             analizedSize+MPD_HDR_SIZE, sizeof(EventMetaInfo::evNum));
            stat.lastEvNum = i->evNum;
            [[fallthrough]];
        case MPD_SYNC_EOB:
        case MPD_SYNC_EOB_TLV:
        case MPD_SYNC_JSON_TLV:
            i->type = mpdHdr.sync;
            stat.lastSyncWord = mpdHdr.sync;
            break;
        case MPD_SYNC_RUN_START:
        case MPD_SYNC_RUN_STOP:
        case MPD_SYNC_FILE_BEGIN:
        case MPD_SYNC_FILE_END:
        {
            int fullSize = MPD_HDR_SIZE+static_cast<int>(mpdHdr.length);
            if(!checkOrWaitFullRec(fullSize))
                break;

            //Trash incoming Start/Stop records
            if(lastMetaInfo == nullptr) {
                const int part1 = qMin(fullSize, byteRB->getContinuosTrashSize());
                byteRB->trashData(part1);
                byteRB->freeData(part1);
                if(fullSize != part1) {
                    const int part2 = fullSize-part1;
                    byteRB->trashData(part2);
                    byteRB->freeData(part2);
                }
            } else {
                lastMetaInfo->skipData += fullSize;
                analizedSize += fullSize;
            }
            stat.totalSize += static_cast<quint64>(fullSize);
            stat.lastSyncWord = mpdHdr.sync;
            continue;
        }
        default:
            if(!badSync) {
                ++stat.badSync;
                qWarning() << objectName() << QString("::analize() bad sync; analizedSize=%1 ; sync=0x%2")
                              .arg(analizedSize).arg(mpdHdr.sync, 8, 16, QChar('0'));
            }
            badSync = true;
            // todo search in local loop
            if(lastMetaInfo == nullptr) {
                byteRB->trashData(1);
                byteRB->freeData(1);
            } else {
                ++(lastMetaInfo->skipData);
                ++analizedSize;
            }
            ++stat.corruptedSize;
            ++stat.totalSize;
            continue;
        }
        if(badSync)
            badSync = false;

        //MPD header is valid wait for full event
        int fullSize = MPD_HDR_SIZE+static_cast<int>(mpdHdr.length);
        if(mpdHdr.sync == MPD_SYNC_EOB || mpdHdr.sync == MPD_SYNC_NORM)
            fullSize += sizeof(MpdEventHdr::evNum);
        if(!checkOrWaitFullRec(fullSize))
            break;

        i->data1 = byteRB->getReadPointer(analizedSize);
        i->dataPos1 = byteRB->getReaderPos(analizedSize);
        i->size1 = qMin(fullSize, byteRB->getContinuosDataSize(analizedSize));
        i->size2 = fullSize-i->size1;
        if(i->size2) {
            i->data2 = byteRB->getReadPointer(analizedSize+i->size1);
        }
        i->free = false;
        i->headerSize = (i->type==MPD_SYNC_NORM || i->type==MPD_SYNC_NORM_TLV || i->type==MPD_SYNC_EOB) ?
                MPD_OLD_HDR_SIZE : MPD_HDR_SIZE;
        lastMetaInfo = i;
//        if(metaInfoBuff->writePos==0)
//            qDebug()<<"first pos evNum="<<mpdHdr.evNum;
        stat.addEvent(static_cast<quint32>(fullSize));
        stat.totalSize += fullSize;
        if(stat.period.isNextPeriod(i->createTime)){
            emit evSizePeriodUpdated(index, stat.period);
            stat.period.newPeriod();
        }
        analizedSize += fullSize;
//        log(QString("::analize() added event pos=%1 evNum=%2")
//            .arg(eventRB->getWriterPos()).arg(mpdHdr.evNum));

        eventRB->writeData();
    }

    runTimer->start();
}

void Analizer::sentStat() const
{
    emit statUpdated(stat);
}

void Analizer::doStateAction(const FsmState &st)
{
    if(performedStateAction == st)
        return;
    qDebug() << objectName() << __PRETTY_FUNCTION__ << st;
    performedStateAction = st;

    switch (st) {
    case FsmState::Configure:
        reset();
        break;
    case FsmState::Run:
        startAnalize();
        break;
    case FsmState::Ready:
        stopAnalize();
        break;
    default:;
    }
}

void Analizer::reset()
{
    qDebug() << objectName() << "start resetting";
    analizedSize=0;

#ifdef RESET_ROUGH
//    qInfo()<<"RESET_ROUGH"<<id<<eventRB->getStrStatus();
    int val = eventRB->getContinuosDataSize(0);
    eventRB->trashData(val);
    val = eventRB->getContinuosDataSize(0);
    if(val)
        eventRB->trashData(val);
    Q_ASSERT(eventRB->getDataSize()==0);
    while(eventRB->getTrashSize()){
        EventMetaInfo *i = eventRB->getClearPointer(0);
        i->free = true;
        eventRB->freeData();
    }
    cleanUpFreedSpace(0);

    Q_ASSERT(eventRB->getFreeSize()==eventRB->getAllSize());
#else
    int pendingBytes = byteRB->getDataSize();
    byteRB->trashData(pendingBytes);
    while(eventRB->getFreeSize() != eventRB->getAllSize()){
        cleanUpFreedSpace(50);
    }
    pendingBytes = byteRB->getDataSize();
    byteRB->trashData(pendingBytes);
#endif


    lastMetaInfo=nullptr;
//    newRawDataWC;
//    wcMutex;
//    allEventsSem; //resetted in Combiner

    Q_ASSERT(!enabled);
//    currentEvent; // uses from Combiner
    stat.clear();
    qDebug() << objectName() << "resetting completed";
}

void Analizer::startAnalize()
{
    qDebug() << objectName() << __PRETTY_FUNCTION__ << "byteRB:" <<byteRB->getStrStatus()
             << "eventRB:" <<eventRB->getStrStatus();

    if(enabled){
        qWarning() << objectName() << "double start";
        return;
    }
    enabled = true;
    runTimer->start();
}

void Analizer::stopAnalize()
{
    if(!enabled){
        qWarning() << objectName() << "double stop";
        return;
    }
    enabled = false;
    runTimer->stop();
}

bool Analizer::cleanUpFreedSpace(int timeoutMs)
{
    bool hasFree = eventRB->getFreeSize();

    int toFree = eventRB->getTrashSize();
    if(hasFree) {
        if(toFree==0)
            return true;
    } else {
        if(toFree==0) {
            // minimize using of timeLeftForRun()
            if(timeoutMs==-1)
                timeoutMs = timeLeftForRun();

            if(!eventRB->waitTrashSpace(1, timeoutMs)) {
                // There were no free space and we could not wait for toFreeSpace
                return false;
            }
            toFree = eventRB->getTrashSize();
            Q_ASSERT(toFree>0);
        }
    }

    //We have at least 1 info to free
//    log(QString("::cleanUpFreedSpace() clean-up: %1 events").arg(toFree));
    for(int shift=0; shift<toFree; ++shift) {
        EventMetaInfo *i = eventRB->getClearPointer(0);
//        log(QString("cleanUpFreedSpace event pos=%1 evNum=%2")
//            .arg(eventRB->getReaderPos()).arg(i->evNum));
        byteRB->trashData(i->size1, i->dataPos1);
        byteRB->freeData(i->size1, i->dataPos1);
        analizedSize -= i->size1;
        if(i->size2){
            byteRB->trashData(i->size2, 0);
            byteRB->freeData(i->size2, 0);
            analizedSize -= i->size2;
        }
        if(i->skipData) {
            const int &skipTotal = i->skipData;
            const int skip1 = qMin(skipTotal, byteRB->getContinuosDataSize(0));
            byteRB->trashData(skip1);
            byteRB->freeData(skip1);
            if(skipTotal != skip1) {
                const int skip2 = skipTotal - skip1;
                byteRB->trashData(skip2);
                byteRB->freeData(skip2);
            }
            analizedSize -= skipTotal;
        }
        Q_ASSERT(analizedSize>=0);
        i->free = true;
        if(lastMetaInfo==i)
            lastMetaInfo = nullptr;
//        bool ret = metaInfoBuff->trashSpace.tryAcquire(1, 0);
        eventRB->freeData(1);
//        Q_ASSERT(ret);
    }
//    bool ret = metaInfoBuff->trashSpace.tryAcquire(toFree, 0);
//    metaInfoBuff->freeData(toFree);
//    Q_ASSERT(ret);
    return true;
}

bool Analizer::checkOrWaitFullRec(int fullSize) const
{
    if(analizedSize+fullSize < 0) {
        eventRB->waitTrashSpace(1, timeLeftForRun());
        return false;
    }
    if(byteRB->getDataSize() < analizedSize+fullSize) {
        return byteRB->waitDataSpace(analizedSize+fullSize, timeLeftForRun());
    }
    return true;
}
