#ifndef COMBINER_H
#define COMBINER_H

#include <QMap>
#include <QObject>
#include <QSet>

#include "base/RcRunId.h"
#include "base/RunDocument.h"
#include "fsm/FsmState.h"
#include "lib-common/MpdRawTypes.h"

#include "CombinerFsm.h"
#include "CombinerStats.h"
#include "EventMetaInfo.h"
#include "EvBConfig.h"
#include "params.h"
#include "OutputTcpServer.h"

class  Analizer;
struct ComplexTlvRecord;
class  FileWriter;
struct FileWriterStats;
template<class T> class RingBuffer;

class QSemaphore;
class QTimer;
class OutputTcpServer;

typedef QPair<quint64, quint64> EventsRange;

struct ClientCombinerData
{
    void removeEv();

    ClientIndex index;
    EventMetaInfo* evMI; // Known pending events
    RingBuffer<EventMetaInfo> *rb; // RB with EventMetaInfo from Analyzer
    int rbDataSpace = 0; // cache of size of available EventMetaInfo in RB
    QMap<quint64, EventsRange> trashStat;
    static const int SuppresLimitMax = 5;
    int suppresLimit = SuppresLimitMax;
};

class Combiner : public CombinerFsm
{
    Q_OBJECT
public:
    explicit Combiner();
    ~Combiner() override;
    void initOutputBuffers(const EvbCombinerConfig &cfg);

signals:
    void status(const QString &) const;
    void statusRaw(const CombinerStats &) const;
//    void clientStateUpdated(ClientIndex index, FsmState state); // from combiner and clients
    void fsmStateChanged(FsmState state) const;
    void error(const QString &) const;
    void startTcpServer(int index, bool state) const;
    void deleteTcp(int) const;
    void resetOutputs() const;
    void startRun(const RcRunId &runId,
                  std::optional<int> writerIndex,
                  std::optional<int> splitId) const;
    void startingTcpWork() const;
    void closeRun() const;
    void closeTcpWork(uint) const;
    void deleteFileWriter() const;
    void fileWriterStatsUpdated(const FileWriterStats &s) const;
    void outputBufferSizeUpdated(int rbId, quint64 size) const;
    void setMonitorPortNum(quint16 port) const;
    void requestForFileSplit(uint id);
    void evSizePeriodUpdated(const EventSizePeriod &s) const;
    void infOutputUpdated(bool en) const;

public slots:
    void onFsmStateChanged(const FsmState &cur, const FsmState &prev);
    void setRequestState(FsmState state, const QVariant &arg);
    void destroyClient(const ClientIndex &clientId);
    void updateAnalizerList(const ClientIndexSet &clients, const ClientIndexSet &nonEvBasedClients,
                            const QVector<Analizer*> &analizers);
    void gotRunDocument(const RunDocument &rd);

signals: void combinerConfigChanged(const EvbCombinerConfig &) const;
public slots: void setCombinerConfig(const EvbCombinerConfig &cfg);

signals: void evNumberConfigChanged(const EvbEvNumberConfig &) const;
public slots: void setEvbEvNumberConfig(const EvbEvNumberConfig &cfg);

signals: void fileSplitterConfigChanged(const EvbFileSplitterConfig &) const;
public slots: void setFileSplitterConfig(const EvbFileSplitterConfig &newConfig);

private slots:
    void init();
    void workLoop();
    void sendStatus();
    void tcpStatusUpdated(const OutputTcpStatus &s);
    void tcpFinished(int tcpId, uint fId);
    void writerFileOpened(bool opened, const RcRunId &_runId);
    void fileSplitted(uint id);
    void outputError(int index, const QString &errStr);
    void onFileWriterStatsUpdated(const FileWriterStats &s);
    void onStoppingAction();

protected:
    virtual void doReqStatePreparations() override;

private:
    void processInit();
    void reset();
    void flushSingleRecord(ClientCombinerData *ccd);
    void flushEvent();
    QList<RingBuffer<char> *> getOutBuffersForRecord(quint32 recLen,
                                                     bool forceFirstOut,
                                                     bool *ok);
    void sendClientPayload(const QList<RingBuffer<char> *> &outBuffers, const EventMetaInfo *ev);
    void writeData(const QList<RingBuffer<char> *> &outBuffers, const char *src, int size);
    void writeData(RingBuffer<char> *buffer, const char *src, int size);
    bool updateClientsEventInfo(ClientCombinerData *ccd, int timeout=0);
    void restoreRealEvNum(const ClientIndex &index, EventMetaInfo* ev);
    int addTcpServer();
    bool writeRunDocument();
    bool writeStartStopRecrd(bool isStart);
    bool writeFileSplitRecrd(bool isStart);
    bool writeComplexRecrd(const ComplexTlvRecord &r);
    void addTrashStat(ClientCombinerData *ccd, quint32 evNum);
    void resizeBuffer();
    void checkStopEvNum();
    void onStartingAction();
    void updateClientSizeCache();
    bool isRunRecWritten();
    bool checkFileSplitting();
    bool checkTimeSplit();
    bool checkEvNumSplit();
    void sendSplitRequest();
    std::optional<int> getWriterIndex() const;
    void resetEventIncomplet();

    ClientIndexSet clientIds; // client id by index of enabled client
    ClientIndexSet nonEvBasedClients;
    ClientIndexSet evBasedClients;
    int clientSizeCache=0;
    QMap<ClientIndex, ClientCombinerData> ccds; // All data coresponding to single Client
    bool initFlag = true;
    bool startRecordWritten = false;
    bool runDocumentWritten = false;
    bool enabled = false;
    EvbEvNumberConfig evNumberCfg;
    quint32 expectedEvNum = evNumberCfg.startEvNum;
    CombinerStats stats;
    RcRunId runId;

    QTimer * const workLoopTimer;
    QTimer * const statusTimer;
    QMap<int, OutputTcpServer *> outputTcpServers;
    QMap<int, QThread *> outputTcpServerThreads;
    OutputTcpServer * monitorTcpServer;
    RingBuffer<char> * const writerRingBuf;
    FileWriter * const fileWriter;
    QThread * const fileWriterThread;
    QByteArray outputEvent;
    bool noFreeSpaceError = false; // for single error string
    uint eventsWithDebug = 0;
    uint finishId = 1;
//    bool prevEventIncomplet;
    int evIncompletCnt = 0;
    QElapsedTimer updateClientsTimer;
    int suppressLogLagging = 0;

    bool waitingResize = false;
    int preferedBufSize = 0;

    // Config
    bool outToFile = true;
    bool outToTcpServer = false;
    EvbCombinerConfig combCfg;
    EvbFileSplitterConfig splitConfig;

    // Pre-allocated fields for local work
    MpdHdr mpdHdr;
    MpdEventHdr mpdEventHdr;

    bool warnEvNumStepFitFail=true;
    bool warnEvNumShift=true;

    QElapsedTimer fileTimer;
    uint fileSplitId = 0;
    bool splitInProgress = false;
    uint writtenEvNum = 0;
    std::optional<RunDocument> runDoc;
};

#endif // COMBINER_H
