#include "FileWriter.h"
#include "metric-sender/MetricSender.h"
#include "metric-sender/MetricTypes.h"
#include "base/ConfigIndexName.h"
#include "base/RcRunId.h"

#include <QFile>
#include <QDir>
#include <QDateTime>
#include <QTimer>

namespace {
const int WRITER_TIMER_PERIOD = 100; //ms
const int METRICS_SEND_PERIOD_MS = 1000;
}

FileWriter::FileWriter(RingBuffer<char> * const writerRB, QObject *parent)
    : QObject(parent)
    , dataFile(new QFile(this))
    , ringBuf(writerRB)
    , stats(new TcpServerStats())
    , writeTimer(new QTimer(this))
    , statusTimer(new QTimer(this))
{
    qRegisterMetaType<EvbWriterConfig>();
    qRegisterMetaType<FileWriterStats>();
    qRegisterMetaType<std::optional<int>>();
    qInfo() << QString("Predefined storage is:").append(config.storageDir);
    connect(writeTimer, &QTimer::timeout, this, &FileWriter::writeData);
    writeTimer->setSingleShot(false);
    writeTimer->start(WRITER_TIMER_PERIOD);
    updateStartTimeStamp();
    defineFileName();
    connect(statusTimer, &QTimer::timeout, this, &FileWriter::sendStatus);
    statusTimer->start(500);

    QTimer * sendMetricsTimer = new QTimer(this);
    connect(sendMetricsTimer, &QTimer::timeout, this, &FileWriter::sendMetrics);
    sendMetricsTimer->start(METRICS_SEND_PERIOD_MS);
}

FileWriter::~FileWriter()
{
    writeTimer->stop();
    statusTimer->stop();
    if(ringBuf->getDataSize()){
        qWarning() << "Some data haven't been written to file";
    }
    if(dataFile->isOpen())
        closeRun();
}

void FileWriter::startRun(const RcRunId &_runId, std::optional<int> _writerIndex, std::optional<int> _splitIndex)
{
    runId = _runId;
    splitIndex = _splitIndex;
    writerIndex = _writerIndex;
    totalStats.writtenTotalSize = 0;

    updateStartTimeStamp();

    if(openFileToWrite())
        emit fileOpened(true, runId);
    sendStatus();
}

void FileWriter::writeData()
{
    writeLatencyElapsedTimer.restart();

    int writeLoops = 0;
    static constexpr const int MaxWritePart = 8e6; // 8MB
    do {
        int size = ringBuf->getContinuosDataSize(0);
        size = std::min(size, MaxWritePart);
        if(size==0)
            break;
        if(!dataFile->isOpen()) {
            const QString errStr = QString("The file isn't open (%1)").arg(dataFile->fileName());
            qWarning() << errStr;
            dropRingBuffer();
            emit error(-1, errStr);
            return;
        }
        qint64 written = dataFile->write(ringBuf->getReadPointer(), size);
        if(written == -1){
            const QString errStr = QString("Error on file write:").append(dataFile->errorString());
            qWarning() << errStr;
            emit error(-1, errStr);
            return;
        }
        ++writeLoops;
//        qInfo() << "written:" << written << "size:" << size;
        ringBuf->trashData(written);
        int s = ringBuf->getContinuosTrashSize();
        ringBuf->freeData(s);
        s = ringBuf->getContinuosTrashSize();
        if(s)
            ringBuf->freeData(s);
        stats->bytesSent += written;
        totalStats.writtenFileSize += written;
        totalStats.writtenTotalSize += written;

        if(written != size){
            if(written == -1){
                totalStats.missedFileSize += size;
                const QString &errStr = QString("Error on file write: %1")
                        .arg(dataFile->errorString());
                qWarning() << errStr;
                emit error(-1, errStr);
            } else {
                totalStats.missedFileSize += size-written;
                dataFile->error();
                const QString &errStr = QString("Incomplete file write: wrote %1 bytes of %2")
                        .arg(written).arg(size).arg(dataFile->errorString() );
                qWarning() << errStr;
                emit error(-1, errStr);
            }
        }
    } while(writeLatencyElapsedTimer.elapsed() < WRITER_TIMER_PERIOD);

    if(ringBuf->getDataSize())
        writeTimer->start(0);
    else
        writeTimer->start(WRITER_TIMER_PERIOD);

    if(writeLoops) {
        double elapsedAver = writeLatencyElapsedTimer.nsecsElapsed();
        elapsedAver /= writeLoops;
        do {
            stats->addLatencyValue(elapsedAver);
        } while(--writeLoops);
    }
}

void FileWriter::closeRun()
{
    closeFile();
    emit fileOpened(false, runId);
    runId = RcRunId();
    splitIndex.reset();
}

void FileWriter::requestForSplit(uint id)
{
    splitIndex = id;
    closeFile();
    if(openFileToWrite()) {
        emit fileSplitted(splitIndex.value());
    }
}

void FileWriter::setWriterConfig(const EvbWriterConfig &cfg)
{
    if(config == cfg)
        return;
    if(config.storageDir != cfg.storageDir)
        qInfo() << QString("Change storage folder to:").append(cfg.storageDir);
    if(config.addLocalProgramIndex != cfg.addLocalProgramIndex)
        qInfo() << QString("Output file name will be with%1 local ProgamIndex")
                   .arg(cfg.addLocalProgramIndex ? "" : "out");
    config = cfg;
    totalStats.predefStorage = config.storageDir;
    defineFileName();
    sendStatus();
}

void FileWriter::configSelectorChanged(const ConfigIndexName &cin)
{
    programIndex = cin.getIndex();
}

void FileWriter::clear()
{
    totalStats.clear();
    totalStats.predefStorage = config.storageDir;
    ringBuf->clear();
    sendMetrics();
    stats->clear();
}

void FileWriter::sendStatus()
{
    totalStats.timeStamp.start();
    totalStats.rbSize = ringBuf->getAllSize();
    totalStats.dataPart = ringBuf->getDataPart();
    emit statusUpdated(totalStats);
}

void FileWriter::sendMetrics()
{
    MetricFields fields{{"free", ringBuf->getFreePart()},
                        {"data", ringBuf->getDataPart()},
                        {"trash", ringBuf->getTrashPart()}};
    MetricTags tags{{"writer_index", writerIndex.has_value()
                    ? QString::number(*writerIndex) : QString()}};
    MetricSender::send("evb.out.file.bytes", stats->bytesSent, tags);
    MetricSender::send("evb.out.file.buffer", fields, tags);
    MetricSender::send("evb.out.file.latency.mean", stats->getMeanLatency(), tags);
}

bool FileWriter::openFileToWrite()
{
    QString &fileName = totalStats.fileName;
    defineFileName();

    QDir dir(QFileInfo(fileName).dir());
    if(!dir.exists())
        dir.mkpath(".");

    if(dataFile->isOpen())
        dataFile->close();
    dataFile->setFileName(fileName);

    totalStats.writtenFileSize = 0;
    if(dataFile->open(QFile::WriteOnly | QFile::NewOnly | QFile::Unbuffered)) {
//        totalStats.missedFileSize = 0;
//        totalStats.writtenFileSize = 0;
        if(config.protectFile) {
            bool ok = dataFile->setPermissions(QFile::WriteOwner);
            if(ok) {
                qInfo() << QFileInfo(*dataFile).absoluteFilePath().prepend("Start writting to file:");
            } else {
                const QString errStr = QFileInfo(*dataFile).absoluteFilePath().prepend("Can't close permissions to file:");
                qWarning() << errStr;
            }
        }
        return true;
    } else {
        QString errStr = dataFile->errorString().prepend("Error on file open:");
        emit error(-1, errStr);
        fileName="";
        qWarning() << errStr;
        return false;
    }
}

void FileWriter::defineFileName()
{
    QString path;
    if(!config.storageDir.isEmpty())
        path = config.storageDir + QDir::separator();

    QStringList fnParts;
    if(!config.filePrefix.isEmpty())
        fnParts << config.filePrefix;
    if(runId.isValid()) {
        if(config.addRunIndex)
            fnParts << runId.runIndex;
        if(config.addRunNumber)
            fnParts << QString("%1").arg(runId.runNumber, 3, 10, QChar('0'));
    }
    if(config.addLocalProgramIndex && !programIndex.toString().isEmpty())
        fnParts << programIndex.toString();
    if(config.addTimestamp)
        fnParts << startTimeStamp;

    if(writerIndex.has_value())
        fnParts <<  QString("ev%1").arg(writerIndex.value());
    if(splitIndex.has_value())
        fnParts << QString("p%1").arg(*splitIndex);

    QString &fn = totalStats.fileName = path + fnParts.join('_') + ".data";
    QDir dir(QFileInfo(fn).dir());
    if(!dir.exists())
        dir.mkpath(".");
    if(!dir.exists()) {
        QString errStr = QString("Output folder '%1' doesn't exist").arg(dir.path());
        qCritical() << errStr;
        emit error(-1, errStr);
    } else {
        qInfo() << QString("Output folder set to '%1'").arg(dir.path());
    }
}

void FileWriter::closeFile()
{
    //    ringBuf->clear();
    if(ringBuf->getFreeSize() != ringBuf->getAllSize()) {
        writeData();
        if(ringBuf->getFreeSize() != ringBuf->getAllSize())
            qWarning() << "closeFile data size:" << ringBuf->getDataSize();
        dropRingBuffer();
    }
    if(!dataFile->flush())
        qWarning() << dataFile->errorString().prepend("Error on file flush:");
    dataFile->close();
    bool ok = true;
    if(config.protectFile) {
        ok = dataFile->setPermissions(QFile::ReadOwner | QFile::ReadUser |
                                           QFile::ReadOther | QFile::ReadGroup);
    }

    if(ok) {
        qInfo() << "File closed";
    } else {
        const QString errStr = "Failed set read permissions on file";
        qWarning() << errStr;
    }
}

void FileWriter::dropRingBuffer()
{
    while(ringBuf->getDataSize()) {
        int s = ringBuf->getContinuosDataSize(0);
        ringBuf->trashData(s);
        qWarning("%s: drop %d bytes", __PRETTY_FUNCTION__, s);
    }
    while(ringBuf->getTrashSize()) {
        int s = ringBuf->getContinuosTrashSize();
        ringBuf->freeData(s);
    }
}

void FileWriter::updateStartTimeStamp()
{
    startTimeStamp = QDateTime::currentDateTime().toString("yyyyMMdd_hhmmss");
}
