#include "SlowControlServer.h"

#include <QTcpServer>
#include <QTimer>
#include <QtNetwork>

#include "SlowControlServer_p.h"
#include "TrigStatus.h"

SlowControlServer::SlowControlServer(QObject *parent) :
    QObject(parent),
    d_ptr(new SlowControlServerPrivate()),
    qTcpServer(new QTcpServer(this))
{
    setObjectName(__func__);
    responce();

#ifdef QT_DEBUG
    QTimer *periodicResponceTimer = new QTimer(this);
    connect(periodicResponceTimer, &QTimer::timeout,
            this, &SlowControlServer::responce);
    periodicResponceTimer->start(1000);
#endif
}

SlowControlServer::~SlowControlServer()
{
    blockSignals(true);
    startListen(false);
    delete qTcpServer;
    blockSignals(false);
}

void SlowControlServer::setSlowControlConfig(const SlowControlConfig &newConfig)
{
    Q_D(SlowControlServer);

    if(d->config.has_value()) {
        if(d->config.value() == newConfig)
            return;
    }
    d->config = newConfig;

    startListen(false);
    if(newConfig.enabled)
        startListen(true);
}

void SlowControlServer::startListen(bool listen)
{
    listening = listen;
    if(listening) {
        init();
    } else {
        qTcpServer->close();
        if(tcpSocket){
            tcpSocket->abort();
            delete tcpSocket;
            tcpSocket = nullptr;
            portInUse=false;
        }
        qDebug() << objectName() << QString(" closed on %1:%2").arg(qTcpServer->serverAddress().toString()).arg(tcpPort);
    }
    updateStatus();
}

void SlowControlServer::updateConnections()
{
    while(QTcpSocket *pendingConnection = qTcpServer->nextPendingConnection()) {
        if(tcpSocket && (!tcpSocket->isValid() || tcpSocket->state() != QTcpSocket::ConnectedState)){
            tcpSocket->abort();
            delete tcpSocket;
            tcpSocket = nullptr;
            portInUse=false;
        }
        if(!tcpSocket){
            tcpSocket = pendingConnection;
            portInUse=true;
            connect(tcpSocket, &QAbstractSocket::readyRead,
                    this, &SlowControlServer::readCommandRequest);
            connect(tcpSocket, &QAbstractSocket::disconnected,
                    this, &SlowControlServer::onDisconnected);
            QHostInfo info = QHostInfo::fromName(tcpSocket->peerAddress().toString());
            QString msg = QString("Accepted connection from %1:%2")
                    .arg(info.hostName())
                    .arg(tcpSocket->peerPort());
            qInfo() << objectName() << msg;
            updateStatus();
#ifdef QT_DEBUG
responce();
#endif
        } else {
            pendingConnection->abort();
            delete pendingConnection;
        }
    }
    if(!tcpSocket) {
        qDebug() << objectName() << " there are no connections.";
    }
}

void SlowControlServer::readCommandRequest()
{
    if(!tcpSocket)
        return;
    while(tcpSocket->bytesAvailable()){
        QByteArray lineData = tcpSocket->readLine();
        QString line = QString::fromLatin1(lineData);
        if(line.startsWith("a")){
            responce();
        }
    }
}

void SlowControlServer::onDisconnected()
{
    portInUse=false;
    updateStatus();
    qInfo() << objectName() << "disconnected";
}

void SlowControlServer::slowControlStateUpdated(QString _state, QString _status)
{
    Q_D(SlowControlServer);
    d->daqState = _state;
    d->daqStatus = _status;
#ifdef QT_DEBUG
    responce();
#endif
}

void SlowControlServer::trigStatusUpdated(const TrigStatus &status)
{
    Q_D(SlowControlServer);
    d->rcState = (status.conn.isReady()) ? SC_STATE_ON : SC_STATE_FAULT;
    d->rcDescr = status.getDeviceIdent();
    if (status.spillStatus.has_value())
        d->spillCount = status.spillStatus.value().count;
    d->evNumber = status.eventNumber();
#ifdef QT_DEBUG
    responce();
#endif
}

void SlowControlServer::spillPlotUpdated(const SpillPlotData &spillPlot)
{
    Q_D(SlowControlServer);
    d->lastSpillPlot = spillPlot;
}

void SlowControlServer::triggerHistogramUpdated(const TriggerHistogramData &trigData)
{
    Q_D(SlowControlServer);
    d->lastTrigData = trigData.histograms.isEmpty() ?
                TriggerHistogramChannelData() :
                trigData.histograms.last(); // fbp.multihist ? [15] : [0]
}

void SlowControlServer::onRunStart(const RcRunId &runId)
{
    Q_D(SlowControlServer);
    d->runId.set(runId);
}

void SlowControlServer::init()
{
    Q_D(SlowControlServer);
    Q_ASSERT(d->config.has_value());
    if(!d->config.has_value())
        return;
    auto &config = d->config.value();

    if(!qTcpServer->isListening()) {
        if(!qTcpServer->listen(QHostAddress::Any, config.port)) {
            QString s = tr("Unable to start on port %1: %2").arg(config.port)
                    .arg(qTcpServer->errorString());
            qCritical() << objectName() << s;
            throw std::runtime_error(s.toStdString());
        }
    }
    tcpPort = qTcpServer->serverPort();

    connect(qTcpServer, &QTcpServer::newConnection,
            this, &SlowControlServer::updateConnections);
    qInfo() << objectName() << QString("listening on %1:%2").arg(qTcpServer->serverAddress().toString()).arg(tcpPort);
}

void SlowControlServer::updateStatus() const
{
    emit connectionStatusUpdated(tcpPort, listening, portInUse);
}

void SlowControlServer::responce() const
{
    Q_D(const SlowControlServer);

    if(!tcpSocket)
        return;

    tcpSocket->write(d->constructResponse());
}
