/*
**    Copyright 2020 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "TrigStatisticsWriter.h"

#include <QDateTime>
#include <QDebug>
#include <QJsonDocument>
#include <QSettings>

#include "base/DaqModule.h"
#include "base/DeviceIndex.h"
#include "message-bus/MessagePublisher.h"
#include "TrigSpillStatisticsConverter.h"

constexpr const char k_state[] = "state";
constexpr const char k_spill[] = "spill";
constexpr const char k_dev[] = "dev";

void TrigStatisticsWriter::writeFile() const
{
    if (!enable)
        return;
    const auto &spillOpt = trigSpillStatistics.status.spillStatus;
    if (!spillOpt.has_value())
        return;
    const auto &spill = spillOpt.value();
    // write file
    QString fileName = QString("rc_%1_%2.txt")
            .arg(QDateTime::currentDateTime().toString("yyyyMMdd_hhmmss"))
            .arg(spill.count);

    QSettings settings(dirName + "/" + fileName, QSettings::IniFormat);
    settings.beginGroup("main");
    settings.setValue("ident", trigSpillStatistics.status.getDeviceIdent());
    settings.setValue("spillCount", spill.count);
    settings.setValue("SpillTimeMs", spill.curSpillTimeMs);
    settings.endGroup();
    settings.beginGroup("spill_plot");
    if (!trigSpillStatistics.spd.data.isEmpty()) {
        settings.setValue("stepSec", trigSpillStatistics.spd.stepSec);
        settings.setValue("startOffset", trigSpillStatistics.spd.startOffset);
        settings.beginGroup("0");
        QStringList sl;
        Q_FOREACH(quint32 val, trigSpillStatistics.spd.data)
            sl += QString("%1").arg(val);
        settings.setValue("data", sl.join(","));
        settings.endGroup();
    }
    settings.endGroup();
    settings.beginGroup("histograms");
    Q_FOREACH(const TriggerHistogramChannelData &thcd, trigSpillStatistics.thd.histograms) {
        if (thcd.isEmpty())
            continue;
        settings.beginGroup(QString("%1").arg(thcd.channel));
        QStringList sl;
        Q_FOREACH(quint32 val, thcd.data)
            sl += QString("%1").arg(val);
        settings.setValue("data", sl.join(","));
        settings.setValue("overflowCount", QString("%1").arg(thcd.overflowCount));
        settings.endGroup();
    }
    settings.endGroup();
    settings.sync();
    if (settings.status() != QSettings::NoError) {
        qWarning() << QString("Error writing status file '%1': %2").arg(settings.fileName()).arg(settings.status());
    }
}
void TrigStatisticsWriter::setStatusDumpEnable(bool val)
{
    enable = val;
}

void TrigStatisticsWriter::setStatusDumpDirName(const QString &str)
{
    dirName = str;
}

void TrigStatisticsWriter::writeStatisticsDump(const TrigSpillStatistics &data)
{
    const TrigStatus &newStatus = data.status;
    const SpillPlotData &newSpd = data.spd;
    const TriggerHistogramData &newThd = data.thd;

    const auto &spillOpt = trigSpillStatistics.status.spillStatus;
    if (!spillOpt.has_value())
        return;
    const auto &spill = spillOpt.value();

    const auto &newSpillOpt = newStatus.spillStatus;
    if (!newSpillOpt.has_value())
        return;
    const auto &newSpill = newSpillOpt.value();

    bool endOfReadout = false;
    // graphs are cleared on beginning of phase 2
    if (spill.status.reg.bit.phase == 4) {
        if (newSpill.status.reg.bit.phase < 2 || newSpill.status.reg.bit.phase == 5) {
            // flush existing graphs
            trigSpillStatistics.spd = newSpd;
            trigSpillStatistics.thd = newThd;
            endOfReadout = true;
        }
        if (newSpill.status.reg.bit.phase == 4) {
            if (spill.count != newSpill.count) {
                endOfReadout = true; // spill incremented
            }
        } else {
            endOfReadout = true;
        }
    }
    if (endOfReadout) {
        writeFile();
    }
    trigSpillStatistics.status = newStatus;
    trigSpillStatistics.spd = newSpd;
    trigSpillStatistics.thd = newThd;
}
