/*
**    Copyright 2015-2021 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "SyslogMessageSender.h"

#include <QCoreApplication>
#include <QString>
#include <QRegularExpression>

#include <iostream>
#include <cstdlib>
#include <QThread>
#include <QMap>
#include <QRegularExpression>
#include <QDebug>

#ifdef Q_OS_UNIX
#include <syslog.h>
#else
#warning "syslog available on UNIX platform only"
#define	LOG_EMERG	1
#define	LOG_ALERT	1
#define	LOG_CRIT	1
#define	LOG_ERR		4
#define	LOG_WARNING	5
#define	LOG_NOTICE	6
#define	LOG_INFO	6
#define	LOG_DEBUG	6

#define	LOG_KERN	(0<<3)
#define	LOG_USER	(1<<3)
#define	LOG_MAIL	(2<<3)
#define	LOG_DAEMON	(3<<3)
#define	LOG_AUTH	(4<<3)
#define	LOG_SYSLOG	(5<<3)
#define	LOG_LPR		(6<<3)
#define	LOG_NEWS	(7<<3)
#define	LOG_UUCP	(8<<3)
#define	LOG_CRON	(9<<3)
#define	LOG_AUTHPRIV	(10<<3)

#define LOG_PID         0x01	/* log the pid with each message */
#define LOG_CONS        0x02	/* log on the console if errors in sending */
#define LOG_ODELAY      0x04	/* delay open until first syslog() (default) */
#define LOG_NDELAY      0x08	/* don't delay open */
#define LOG_NOWAIT      0x10	/* don't wait for console forks: DEPRECATED */
#define LOG_PERROR      0x20	/* log to stderr as well */

void closelog(void) {}
void openlog(const char *, int, int) {}
void syslog(int, const char *, ...) {}
#endif

namespace {

QMap<QString, int> syslog_prio_map_init()
{
    QMap<QString, int> syslog_prio_map;
    syslog_prio_map["emerg"]   = LOG_EMERG;
    syslog_prio_map["alert"]   = LOG_ALERT;
    syslog_prio_map["crit"]    = LOG_CRIT;
    syslog_prio_map["warning"] = LOG_WARNING;
    syslog_prio_map["notice"]  = LOG_NOTICE;
    syslog_prio_map["info"]    = LOG_INFO;
    syslog_prio_map["debug"]   = LOG_DEBUG;
    syslog_prio_map["debug 0"] = LOG_DEBUG;
    syslog_prio_map["debug 1"] = LOG_DEBUG;
    syslog_prio_map["debug 2"] = LOG_DEBUG;
    syslog_prio_map["debug 3"] = LOG_DEBUG;
    syslog_prio_map["debug 4"] = LOG_DEBUG;
    syslog_prio_map["debug 5"] = LOG_DEBUG;
    return syslog_prio_map;
}
}

QMap<QString, int> SyslogMessageSender::syslog_prio_map = syslog_prio_map_init();

QString SyslogMessageSender::trimmedText(const QString &text)
{
    QString newtext = text;
    // strip quotes
    if (newtext.startsWith(QChar('"')) && newtext.endsWith(QChar('"'))) {
        newtext.remove(0, 1);
        newtext.remove(newtext.length()-1, 1);
    }
    // trim whitespace
    newtext = newtext.trimmed();
    return newtext;
}

void SyslogMessageSender::sendSyslogMessage(QtMsgType type, const QMessageLogContext &context, const QString &text)
{
    int priority = 0;
    switch(type) {
    case QtDebugMsg :
        priority = LOG_DEBUG;
        break;
    case QtInfoMsg :
        priority = LOG_INFO;
        break;
    case QtWarningMsg :
        priority = LOG_WARNING;
        break;
    case QtCriticalMsg :
        priority = LOG_CRIT;
        break;
    case QtFatalMsg :
        priority = LOG_ALERT;
        break;
    default:
        priority = LOG_INFO;
    }
    QString newtext = trimmedText(text);
    // format: "[priority] message..."
    if (text.startsWith("[")) {
        QRegularExpression re("^\\[([^]]+)\\]\\s*(.*)");
        QRegularExpressionMatch match = re.match(text);
        if (match.hasMatch()) {
            QString text_priority = match.captured(1);
            QString text_message = match.captured(2);
            if (syslog_prio_map.contains(text_priority)) {
                priority = syslog_prio_map[text_priority];
                newtext = text_message;
            } else {
                // qDebug() << "Unknown log priority tag " + text_priority;
            }
        }
    }
    QString category(context.category);
    if (!category.isEmpty() && category != "default")
        newtext = category + ": " + newtext;
    if (type == QtDebugMsg || type == QtFatalMsg) {
        QString file(context.file);
        QRegularExpression re1("^.*\\/home/[^\\/]+\\/");
        file.remove(re1);
        QRegularExpression re2(".*\\b(libs|apps|src|git|svn)\\/");
        file.remove(re2);
        newtext = file + ":" + QString::number(context.line) + ": " + newtext;
    }
    if (type == QtDebugMsg)
        std::cerr << newtext.toStdString() << std::endl << std::flush;
    else
        log(priority, newtext.toLocal8Bit());
    if (type == QtFatalMsg) {
        qInstallMessageHandler(0);
        std::cerr << std::endl;
        std::cerr << "Fatal error at " << newtext.toStdString() << std::endl << std::endl << std::flush;
        exit(1);
    }
    const QString thrName = QThread::currentThread()->objectName();
    emit instance().logMessage(type, category, text, thrName);
}

SyslogMessageSender::~SyslogMessageSender()
{
    remove();
    destroyed = true;
}

void SyslogMessageSender::install()
{
    qRegisterMetaType<QtMsgType>("QtMsgType");
    const char *ident=strdup(
                QString("afi-" + QCoreApplication::applicationName()
                        .toLower()
                        .replace(QRegularExpression("\\W+"), "_")).toLatin1());
    openlog(ident, LOG_NOWAIT | LOG_PERROR | LOG_PID, LOG_USER);
    if (prevMsgHandler)
        qWarning() << "SyslogMessageSender: already installed";
    auto h = qInstallMessageHandler(sendSyslogMessage);
    if (!prevMsgHandler)
        prevMsgHandler = h;
}

void SyslogMessageSender::remove()
{
    if(prevMsgHandler){
        qInstallMessageHandler(prevMsgHandler);
        prevMsgHandler = nullptr;
    }
}

void SyslogMessageSender::log(int priority, QString message)
{
    syslog(priority, "%s", message.toStdString().c_str());
}
