/*
**    Copyright 2009-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef DSPLIB_DSPLIB_H
#define DSPLIB_DSPLIB_H

#include <cmath>
#include <vector>

namespace dsplib {

/** \brief Ratio between FWHM and sigma of a Gaussian curve
*
* Full Width at Half Maximum / sigma
* equals to sqrt(2*log(2))
*/
constexpr const double GausFWHM = 2.35482004503094933e+00;

inline int BITS2MAXINT(int bits) {return (1<<(bits-1)) - 1;}
inline int BITS2MININT(int bits)  {return -(1<<(bits-1));}
inline unsigned int BITS2MAXUINT(int bits) {return (1<<bits) - 1;}
inline unsigned int BITS2MINUINT(int) {return 0;}

// Find a,b for array of (x,y) of length n
// y = ax + b
void LinearFit (const double *x, const double *y, int n, double *a, double *b, double *c);
void LinearFitInt (const int *y, int n, double *a, double *b, double *c);

/// 2nd derivative
void deriv2 (const int *x, int len, int *out);

/// Return DFT noise floor in dB
double fft_noisefloor (double rmsnoise, int rank);

/// Rectangular to halfcomplex DFT format
std::vector<double> rect2hc(const std::vector<double> &S, const std::vector<double> &C);

/// Halfcomplex to rectangular DFT format
void hc2rect(const std::vector<double> &hc, std::vector<double> *S, std::vector<double> *C);


/// Complex Divide operation in rectangular form: (e+if)=(a+ib)/(c+id)
void Compdiv (double a, double b, double c, double d, double *e, double *f); ///< @deprecated

// Normalize to integral=1
void Normalize (std::vector<double> *x); //! deprecated

/// Scale to fit in <bits>-bit integer
std::vector<int> Make_Discrete (const std::vector<double> &x, int bits);
void Make_Discrete (const double *x, int *out, int n, int bits);  ///< @deprecated

/// Arithmetic sum of array elements (int)
int intsum (const std::vector<int> &x);

///// Max/Min value of array elements (int)
//double maxdbl (const std::vector<double> &x);
//double mindbl (const std::vector<double> &x);
/// DC value
double dcdbl (const std::vector<double> &x);
/// RMS value
double rmsdbl (const std::vector<double> &x, double dc);

class Decibels
{
public:
    /// Decibels to amplitude conversion
    static double db2val(double db) { return exp (db / 20. * log (10.)); }
    /// Amplitude to decibels conversion
    static double val2db(double value) { return 20. * (log(value) / log (10.)); }
};
[[deprecated]] double db2val (double db);
[[deprecated]] double val2db (double val);

/// An unnormalized gaussian function with mean=0 and sigma
///
/// Divide result by sqrt(2*Pi) = 2.50662827463100024 if normalized output needed
double gauss (double x, double sigma);

///  Machine independent random number generator.
///
///  Produces uniformly-distributed floating points in [0,1]
///  Identical sequence on all machines of >= 32 bits.
///  Periodicity > 10**14
double rndm ();

/// returns a uniform deviate on the interval ]x1, x2].
double ranuni (double x1, double x2);

/// Return a number distributed following a gaussian with \a sigma
double rangaus (double sigma);

/// \brief Simulate detector pulse
///
/// \param a amplitude
/// \param t time
/// \param rt risetime
/// \param ft falltime
double detpulse (double a, double t, double rt, double ft);

// Convolution
// size of output is l1 + l2 - 1
// Integer+ShortInteger -> Integer
void convolve_i_s_i (const int *x1, const int l1, const short int *x2, const int l2, int *out);

template <class U, class V, class W>
void convolve(const std::vector<U> &u, const std::vector<V> &v, std::vector<W> *out)
{
    assert(out);
    if (!out)
        return;
    out->clear();
    size_t u_size = u.size();
    size_t v_size = v.size();
    if (u_size == 0 || v.size() == 0)
        return;
    size_t out_size = u_size + v_size - 1;
    out->resize(out_size);
    for (size_t i = 0; i < u_size; i++)
        for (size_t j = 0; j < v_size; j++)
        {
            (*out)[i + j] += u[i] * v[j];
        }
}

} // namespace dsplib

#endif                          /* DSPLIB_DSPLIB_H */
