/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "AsyncDeviceCommander.h"

#include <QDebug>

AsyncDeviceCommander::AsyncDeviceCommander(QObject *parent)
    : QObject{parent}
{
    qRegisterMetaType<DeviceCommand>();
}

void AsyncDeviceCommander::submitCommand(const DeviceCommand &newCommand)
{
    MultiDeviceTask task;
    task.command = newCommand;
    bool broadcast = newCommand.deviceIndex.isNull();
    const auto &enabledDevices = deviceList.getEnabled();
    if (enabledDevices.empty()) {
        completeNoDevices(newCommand);
        return;
    }
    for (const auto &i: enabledDevices) {
        if (broadcast || i == newCommand.deviceIndex) {
            DeviceCommand c(newCommand);
            c.deviceIndex = i;
            task.commands.insert(i, c);
        }
    }
    tasks.insert(newCommand.command.seq, task);
    if (broadcast) {
        emit submitDeviceCommand(newCommand.deviceIndex, newCommand);
    } else {
        for (const auto &t : qAsConst(task.commands)) {
            emit submitDeviceCommand(t.deviceIndex, t);
        }
    }
}

void AsyncDeviceCommander::completeNoDevices(const DeviceCommand &newCommand)
{
    auto result(newCommand);
    result.result.seq = result.command.seq;
    result.result.error = ENODEV;
    qDebug().noquote() << QString("command ignored (no devices)");
    emit commandComplete(result);
}

void AsyncDeviceCommander::deviceCommandComplete(const DeviceIndex &index, const DeviceCommand &completeCommand)
{
    const auto seq = completeCommand.result.seq;
    auto foundTask = tasks.find(seq);
    if (foundTask == tasks.end()) {
        qWarning().noquote() << QString("%1: unknown command sequence: %2").arg(__func__).arg(seq);
        return;
    }
    auto &task = *foundTask;
    const auto deviceIndex = completeCommand.deviceIndex;
    auto foundCommand = task.commands.find(deviceIndex);
    if (foundCommand == task.commands.end()) {
        qWarning().noquote() << QString("%1: unknown device: %2").arg(__func__).arg(deviceIndex.getIdent());
        return;
    }
    task.results[deviceIndex] = completeCommand;
    checkTaskResults(task);
}

void AsyncDeviceCommander::checkTaskResults(const MultiDeviceTask &task)
{
    const auto &taskDevices = task.commands.keys();
    QList<DeviceIndex> wantedDevices;
    for (const auto &index : taskDevices) {
        if (deviceList[index].enabled)
            wantedDevices.push_back(index);
    }
    if (wantedDevices.empty()) {
        completeNoDevices(task.command);
        return;
    }
    for (const auto &k : wantedDevices) {
        if (!task.results.contains(k))
            return;
    }
    if (task.results.empty())
        return;
    QList<int> errors;
    QMap<DeviceIndex, QVariant> values;
    for (const auto &r : task.results) {
        errors.push_back(r.result.error);
        if (!r.result.data.isNull())
            values.insert(r.deviceIndex, r.result.data);
    }
    int min_error = *std::min_element(errors.cbegin(), errors.cend());
    int max_error = *std::max_element(errors.cbegin(), errors.cend());
    int error;
    if (min_error == max_error)
        error = min_error;
    else
        error = EIO;

    auto ret(task.command);
    ret.result.seq = ret.command.seq;
    ret.result.error = error;
    ret.result.data = QVariant::fromValue(values);
    if (error)
        qWarning().noquote() << QString("command failed: %1").arg(strerror(error));
    else
        qDebug().noquote() << QString("command complete");
    emit commandComplete(ret);
}
