/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "NetworkPortModulePrivate.h"

#include "NetworkPortConfig.h"
#include "NetworkPortModule.h"
#include "NetworkPortModuleRegisters.h"
#include "NetworkPortStatus.h"
#include "regio/RegOpStream.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

const uint32_t gtx_phy_csr_magic = 0xc11ba993;
const uint32_t gtx_pcspma_csr_magic = 0x7ae849f9;
const uint32_t mac_csr_v1_magic = 0x8b9a8b2b;
const uint32_t mac_csr_v2_magic = 0x58bc2e51;
const uint32_t hwip_csr_magic = 0xd5393e22;
const uint32_t mstream_csr_magic = 0x4d533278;

template<>
RegOpVector NetworkPortModulePrivate::getRegOp<NetworkPortDescription>() const
{
    RegOpVector r;
    const auto &base = REG_ETH_PORT_DESCR_BASE;

    r.RegRead32(base+REG32_ETH_PORT_DESCR_MAGIC);
    r.RegRead16(base+REG_ETH_PORT_DESCR_CAPABILITY);
    r.RegRead16(base+REG_ETH_PORT_DESCR_PORT_ID);

    r.RegRead16(base+REG_ETH_PORT_DESCR_PHY_REG_BASE);
    r.RegRead16(base+REG_ETH_PORT_DESCR_PCS_PMA_REG_BASE);
    r.RegRead16(base+REG_ETH_PORT_DESCR_MAC_REG_BASE);
    r.RegRead16(base+REG_ETH_PORT_DESCR_HWIP_REG_BASE);
    r.RegRead16(base+REG_ETH_PORT_DESCR_MSTREAM_REG_BASE);
    return r;
}

template<>
bool NetworkPortModulePrivate::isAvailable<PhyPortStatus>() const
{
    return descr.isValid() && descr.capability_reg().phyCsr;
}

template<>
RegOpVector NetworkPortModulePrivate::getRegOp<PhyPortStatus>() const
{
    if(!isAvailable<PhyPortStatus>())
        return {};
    RegOpVector r;
    const auto &base = descr.phyRegBase;
    r.RegRead32  (base+REG32_ETH_PORT_PHY_MAGIC);
    r.RegRead32  (base+REG32_ETH_PORT_PHY_STATUS);
    return r;
}

template<>
bool NetworkPortModulePrivate::isAvailable<PcspmaPortStatus>() const
{
    return descr.isValid() && descr.capability_reg().pcsPmaCsr;
}

template<>
RegOpVector NetworkPortModulePrivate::getRegOp<PcspmaPortStatus>() const
{
    if(!isAvailable<PcspmaPortStatus>())
        return {};
    RegOpVector r;
    const auto &base = descr.pcsPmaRegBase;
    r.RegRead32  (base+REG32_ETH_PORT_PCSPMA_MAGIC);
    r.RegRead32  (base+REG32_ETH_PORT_PCSPMA_STATUS);
    for (int i=0; i<PcspmaPortStatus::counters_size; i++) {
        r.RegRead32  (base+REG32_ETH_PORT_PCSPMA_COUNTER_0 + 2*i);
    }
    for (int i=0; i<PcspmaPortStatus::timers_size; i++) {
        r.RegRead32  (base+REG32_ETH_PORT_PCSPMA_TIMER_0 + 2*i);
    }

    return r;
}

template<>
bool NetworkPortModulePrivate::isAvailable<MacPortStatus>() const
{
    return descr.isValid() && descr.capability_reg().macCsr;
}

template<>
RegOpVector NetworkPortModulePrivate::getRegOp<MacPortStatus>() const
{
    if(!isAvailable<MacPortStatus>())
        return {};
    RegOpVector r;
    const auto &base = descr.macRegBase;
    r.RegRead32  (base+REG32_ETH_PORT_MAC_MAGIC);
    r.RegRead32  (base+REG32_ETH_PORT_MAC_STATUS);
    r.RegRead32  (base+REG32_ETH_PORT_MAC_MTU);
    for (int i=0; i<MacPortStatus::counters_size; i++) {
        r.RegRead32  (base+REG32_ETH_PORT_MAC_COUNTER_0 + 2*i);
    }
    return r;
}

template<>
bool NetworkPortModulePrivate::isAvailable<HwipPortStatus>() const
{
    return descr.isValid() && descr.capability_reg().hwipCsr;
}

template<>
RegOpVector NetworkPortModulePrivate::getRegOp<HwipPortStatus>() const
{
    if(!isAvailable<HwipPortStatus>())
        return {};
    RegOpVector r;
    const auto &base = descr.hwipRegBase;
    r.RegRead32  (base+REG32_ETH_PORT_HWIP_MAGIC);
    r.RegRead32  (base+REG32_ETH_PORT_HWIP_STATUS);
    r.RegRead32  (base+REG32_ETH_PORT_HWIP_HOST_IP);
    r.RegRead64  (base+REG64_ETH_PORT_HWIP_HOST_MAC);
    r.RegRead32  (base+REG32_ETH_PORT_HWIP_REGIO_ERROR_ADDR);
    r.RegRead32  (base+REG32_ETH_PORT_HWIP_REGIO_ERROR_CODE);
    r.RegRead32  (base+REG32_ETH_PORT_HWIP_MLINK_CONFIG);
    r.RegRead32  (base+REG32_ETH_PORT_HWIP_MASTER_IP);
    r.RegRead16    (base+REG_ETH_PORT_HWIP_MASTER_PORT);
    r.RegRead32  (base+REG32_ETH_PORT_HWIP_MSTREAM_IP);
    r.RegRead16    (base+REG_ETH_PORT_HWIP_MSTREAM_PORT);
    for (int i=0; i<HwipPortStatus::counters_size; i++) {
        r.RegRead32  (base+REG32_ETH_PORT_HWIP_COUNTER_0 + 2*i);
    }
    for (int i=0; i<HwipPortStatus::timers_size; i++) {
        r.RegRead32  (base+REG32_ETH_PORT_HWIP_TIMER_0 + 2*i);
    }
    return r;
}

template<>
bool NetworkPortModulePrivate::isAvailable<MstreamPortStatus>() const
{
    return descr.isValid() && descr.capability_reg().mstreamCore;
}

template<>
RegOpVector NetworkPortModulePrivate::getRegOp<MstreamPortStatus>() const
{
    if(!isAvailable<MstreamPortStatus>())
        return {};
    RegOpVector r;
    const auto &base = descr.mstreamRegBase;
    r.RegRead32  (base+REG32_ETH_PORT_MSTREAM_MAGIC);
    r.RegRead32  (base+REG32_ETH_PORT_MSTREAM_STATUS);
    r.RegRead16    (base+REG_ETH_PORT_MSTREAM_MTU_MIN);
    r.RegRead16    (base+REG_ETH_PORT_MSTREAM_MTU_MAX);
    r.RegRead16    (base+REG_ETH_PORT_MSTREAM_MTU);
    r.RegRead16    (base+REG_ETH_PORT_MSTREAM_MAX_PACKET_SIZE);
    r.RegRead32  (base+REG32_ETH_PORT_MSTREAM_DEFRAG_TIMEOUT);
    r.RegRead32  (base+REG32_ETH_PORT_MSTREAM_RETX_TIME);
    r.RegRead32  (base+REG32_ETH_PORT_MSTREAM_RETX_CONFIG);
    r.RegRead32  (base+REG32_ETH_PORT_MSTREAM_RETX_EST_RTT);
    r.RegRead32  (base+REG32_ETH_PORT_MSTREAM_CLK_KHZ);
    for (int i=0; i<MstreamPortStatus::counters_size; i++) {
        r.RegRead32  (base+REG32_ETH_PORT_MSTREAM_COUNTER_0 + 2*i);
    }
    for (int i=0; i<MstreamPortStatus::timers_size; i++) {
        r.RegRead32  (base+REG32_ETH_PORT_MSTREAM_TIMER_0 + 2*i);
    }
    return r;
}

template<>
RegOpVector NetworkPortModulePrivate::getRegOp<NetworkPortStatus>() const
{
    RegOpVector r;
    r << getRegOp<PhyPortStatus>();
    r << getRegOp<PcspmaPortStatus>();
    r << getRegOp<MacPortStatus>();
    r << getRegOp<HwipPortStatus>();
    r << getRegOp<MstreamPortStatus>();
    return r;
}

static bool verbose = false;

RegOpStream &operator>>(RegOpStream &in, NetworkPortDescription &descr)
{
    uint32_t magic = 0;
    in >> magic;
    in >> descr.capability;
    uint16_t id;
    in >> id;
    descr.id = id;

    in >> descr.phyRegBase;
    in >> descr.pcsPmaRegBase;
    in >> descr.macRegBase;
    in >> descr.hwipRegBase;
    in >> descr.mstreamRegBase;

    descr.valid = (magic == AFI_SDB_ETH_PORT_V1);
    if (verbose && !descr.valid && (magic != 0)) {
        qWarning() << QString("Unexpected NetworkPortDescription magic: %1")
                          .arg(magic, 8, 16, QChar('0'));
    }
    return in;
}

RegOpStream &operator>>(RegOpStream &in, PhyPortStatus &phy)
{
    uint32_t magic = 0;
    in >> magic;
    in >> phy.status;

    phy.valid = (magic == gtx_phy_csr_magic);
    if (verbose && !phy.valid && (gtx_phy_csr_magic != 0)) {
        qWarning() << QString("Unexpected gtx_phy_csr_magic: %1")
                          .arg(magic, 8, 16, QChar('0'));
    }
    return in;
}

RegOpStream &operator>>(RegOpStream &in, PcspmaPortStatus &pcspma)
{
    uint32_t magic = 0;
    in >> magic;
    in >> pcspma.status;
    in >> pcspma.counters;
    in >> pcspma.timers;

    pcspma.valid = (magic == gtx_pcspma_csr_magic);
    if (verbose && !pcspma.valid && (gtx_pcspma_csr_magic != 0)) {
        qWarning() << QString("Unexpected gtx_pcspma_csr_magic: %1")
                          .arg(magic, 8, 16, QChar('0'));
    }
    return in;
}

RegOpStream &operator>>(RegOpStream &in, MacPortStatus &mac)
{
    uint32_t magic = 0;
    in >> magic;
    in >> mac.status;
    in >> mac.mtu;
    in >> mac.counters;

    mac.valid = (magic == mac_csr_v2_magic);
    if (verbose && !mac.valid && (mac_csr_v2_magic != 0)) {
        qWarning() << QString("Unexpected mac_csr_magic: %1")
                          .arg(magic, 8, 16, QChar('0'));
    }
    return in;
}

RegOpStream &operator>>(RegOpStream &in, HwipPortStatus &hwip)
{
    uint32_t magic = 0;
    in >> magic;
    in >> hwip.status;
    in >> hwip.host_ip;
    in >> hwip.host_mac;
    in >> hwip.regio_error_addr;
    in >> hwip.regio_error_code;
    in >> hwip.mlink_config;
    in >> hwip.master_ip;
    in >> hwip.master_port;
    in >> hwip.mstream_ip;
    in >> hwip.mstream_port;
    in >> hwip.counters;
    in >> hwip.timers;
    hwip.valid = (magic == hwip_csr_magic);
    if (verbose && !hwip.valid && (magic != 0)) {
        qWarning() << QString("Unexpected hwip_csr_magic: %1")
                          .arg(magic, 8, 16, QChar('0'));
    }
    return in;
}

RegOpStream &operator>>(RegOpStream &in, MstreamPortStatus &mstream)
{
    uint32_t magic = 0;
    in >> magic;
    in >> mstream.status;
    in >> mstream.mtu_min;
    in >> mstream.mtu_max;
    in >> mstream.mtu;
    in >> mstream.max_packet_size;
    in >> mstream.defrag_timeout;
    in >> mstream.retx_ack_time;
    in >> mstream.retx_config;
    in >> mstream.retx_est_rtt;
    in >> mstream.clk_khz;
    in >> mstream.counters;
    in >> mstream.timers;

    mstream.valid = (magic == mstream_csr_magic);
    if (verbose && !mstream.valid && (magic != 0)) {
        qWarning() << QString("Unexpected mstream_csr_magic: %1")
                          .arg(magic, 8, 16, QChar('0'));
    }
    return in;
}

NetworkPortDescription NetworkPortModulePrivate::getDescription(const mlink::RegOpVector &regRes) const
{
    NetworkPortDescription descr;
    RegOpStream in(regRes);
    in >> descr;
    if(!descr.isValid())
        return {};
    return descr;
}

NetworkPortDescription NetworkPortModulePrivate::getDescription_v1() const
{
    NetworkPortDescription descr;
    descr.valid = true;
    descr_cap_reg_t cap{.raw=0};
    cap.phyCsr = 1;
    cap.pcsPmaCsr = 1;
    cap.macCsr = 1;
    cap.hwipCsr = 1;
    cap.mstreamCore = 1;
    descr.capability = cap.raw;

    descr.phyRegBase = REG_ETH_PORT_PHY_BASE;
    descr.pcsPmaRegBase = REG_ETH_PORT_PCSPMA_BASE;
    descr.macRegBase = REG_ETH_PORT_MAC_BASE;
    descr.hwipRegBase = REG_ETH_PORT_HWIP_BASE;
    descr.mstreamRegBase = REG_ETH_PORT_MSTREAM_BASE;

    return descr;
}

NetworkPortStatus NetworkPortModulePrivate::getStatus(const RegOpVector &result) const
{
    RegOpStream in(result);
    NetworkPortStatus status;
    status.descr = descr;
    if(isAvailable<PhyPortStatus>())
        in >> status.phy;
    if(isAvailable<PcspmaPortStatus>())
        in >> status.pcspma;
    if(isAvailable<MacPortStatus>())
        in >> status.mac;
    if(isAvailable<HwipPortStatus>())
        in >> status.hwip;
    if(isAvailable<MstreamPortStatus>())
        in >> status.mstream;
    if (!in.isValid())
        return {};
    return status;
}

bool NetworkPortModulePrivate::setConfig(const NetworkPortConfig &newConfig)
{
    Q_Q(NetworkPortModule);
    config = newConfig;
    config.applyLimits();

    if(!isAvailable<MstreamPortStatus>())
        return true;

    RegOpVector r;
    const auto &base = descr.mstreamRegBase;
    r.RegWrite16(base+REG_ETH_PORT_MSTREAM_MTU, config.mtu);
    r.RegWrite32(base+REG32_ETH_PORT_MSTREAM_DEFRAG_TIMEOUT, config.defragTimeoutUs);
    return q->regOpExecRebased(r);
}
