#include "TqdcCoreConfig.h"

#include <QDataStream>
#include <QDebug>

#include "BoardGpio/BoardGpioConfig.h"
#include "HitCore/HitCoreConfig.h"
#include "TdcCore/TdcCoreConfig.h"
#include "TqdcHardwareDef.h"
#include "WaveRec/WaveRecConfig.h"
#include "WaveTrig/WaveTrigConfig.h"
#include "base/DeviceId.h"

namespace {
static const std::vector<std::pair<QMap<int, bool> TqdcCoreConfig::*, bool>> DEF_LIST_BOOL
{
    {&TqdcCoreConfig::chTdcEn,          TqdcCoreConfig::DEF_TDC_EN},
    {&TqdcCoreConfig::chAdcEn,          TqdcCoreConfig::DEF_ADC_EN},
    {&TqdcCoreConfig::chWfInvert,       TqdcCoreConfig::DEF_INV},
    {&TqdcCoreConfig::chInv,            TqdcCoreConfig::DEF_INV},
    {&TqdcCoreConfig::chTrigEn,         TqdcCoreConfig::DEF_TRIG_EN},
    {&TqdcCoreConfig::chDigTrigEn,      TqdcCoreConfig::DEF_DIG_TRIG_EN},
    {&TqdcCoreConfig::chDigTrigRising,  TqdcCoreConfig::DEF_DIG_TRIG_RISING},
    {&TqdcCoreConfig::chDigZsEn,        TqdcCoreConfig::DEF_ZS_EN},
    {&TqdcCoreConfig::chDigZsRising,    TqdcCoreConfig::DEF_ZS_RISING},
};
static const std::vector<std::pair<QMap<int, int> TqdcCoreConfig::*, int>> DEF_LIST_INT
{
    {&TqdcCoreConfig::chThr,        TqdcCoreConfig::DEF_THR},
    {&TqdcCoreConfig::chDigThr,     TqdcCoreConfig::DEF_DIG_THR},
    {&TqdcCoreConfig::chGain,       TqdcCoreConfig::DEF_GAIN},
    {&TqdcCoreConfig::chDigZsThr,   TqdcCoreConfig::DEF_ZS_THR},
};
}

const bool TqdcCoreConfig::DEF_TDC_EN = true;
const bool TqdcCoreConfig::DEF_ADC_EN  = true;
const bool TqdcCoreConfig::DEF_INV = false;
const int TqdcCoreConfig::DEF_THR = 0;
const bool TqdcCoreConfig::DEF_TRIG_EN = false;
const int TqdcCoreConfig::DEF_DIG_THR = 0;
const bool TqdcCoreConfig::DEF_DIG_TRIG_EN = false;
const bool TqdcCoreConfig::DEF_DIG_TRIG_RISING = true;
const TqdcCoreConfig::Gain TqdcCoreConfig::DEF_GAIN = TqdcCoreConfig::Gain1;
const bool TqdcCoreConfig::DEF_ZS_EN = false;
const bool TqdcCoreConfig::DEF_ZS_RISING = true;
const int TqdcCoreConfig::DEF_ZS_THR = 0;

void TqdcCoreConfig::setField(QMap<int, int> TqdcCoreConfig::*f, int thr)
{
    for (int ch=0; ch<nch; ++ch) {
        QMap<int, int> &field = this->*f;
        field[ch] = thr;
    }
}

bool TqdcCoreConfig::getDefaultFieldVal(QMap<int, bool> TqdcCoreConfig::*f)
{
    for(auto p : DEF_LIST_BOOL)
        if(p.first == f)
            return p.second;

    Q_UNREACHABLE();
    return false;
}

int TqdcCoreConfig::getDefaultFieldVal(QMap<int, int> TqdcCoreConfig::*f)
{
    for(auto p : DEF_LIST_INT)
        if(p.first == f)
            return p.second;

    Q_UNREACHABLE();
    return 0;
}

bool TqdcCoreConfig::hasTdc(DeviceId deviceId)
{
    switch (deviceId) {
    case DEVICE_ID_TDC72VXS:
    case DEVICE_ID_TDC64VHLE:
    case DEVICE_ID_TDC64VLE:
    case DEVICE_ID_TQDC16VS_E:
        return true;
    default: return false;
    }
}

bool TqdcCoreConfig::hasAdc(DeviceId deviceId)
{
    switch (deviceId) {
    case DEVICE_ID_TQDC16VS_E:
    case DEVICE_ID_ADC64:
    case DEVICE_ID_ADC64S2:
    case DEVICE_ID_ADC64WR:
    case DEVICE_ID_ADC64VE:
    case DEVICE_ID_ADC64VE2:
    case DEVICE_ID_ADC64ECAL:
    case DEVICE_ID_ADC64VE_XGE:
        return true;
    default: return false;
    }
}

bool TqdcCoreConfig::hasInvert(DeviceId deviceId)
{
    switch (deviceId) {
    case DEVICE_ID_TQDC16VS_E:
        return true;
    default: return false;
    }
}

bool TqdcCoreConfig::hasTrig(DeviceId deviceId)
{
    return hasAdc(deviceId);
}

int TqdcCoreConfig::getNumberOfChannels(DeviceId deviceId)
{
    switch (deviceId) {
    case DEVICE_ID_BPM4100:
        return 4;
    case DEVICE_ID_8CADC:
    case DEVICE_ID_ADC8E:
    case DEVICE_ID_EVADC:
        return 8;
    case DEVICE_ID_ADC16LTCM:
    case DEVICE_ID_ADCM16:
        //    case DEVICE_ID_ADC16LTCS: return 16;
    case DEVICE_ID_ADCM163:
    case DEVICE_ID_TQDC16VS_E:
        return 16;
    case DEVICE_ID_ADC32P:
        return 32;
    case DEVICE_ID_ADC64:
    case DEVICE_ID_ADC64S2:
    case DEVICE_ID_ADC64WR:
    case DEVICE_ID_ADC64VE:
    case DEVICE_ID_ADC64ECAL:
    case DEVICE_ID_ADC64VE_XGE:
    case DEVICE_ID_ADC64VE2:
    case DEVICE_ID_TDC64VHLE:
    case DEVICE_ID_TDC64VLE:
        return 64;
    case DEVICE_ID_TDC72VXS:
        return 72;
    default: return 1;
    }
}

void TqdcCoreConfig::setNCh(int n)
{
    for(int ch=nch; ch<n; ++ch) {
        for(auto p : DEF_LIST_INT) {
            auto &map = this->*(p.first);
            map[ch] = map.value(ch, p.second);
        }
        for(auto p : DEF_LIST_BOOL) {
            auto &map = this->*(p.first);
            map[ch] = map.value(ch, p.second);
        }
    }
    nch = n;
}

bool TqdcCoreConfig::isGain4() const
{
    return !chGain.isEmpty() && chGain.first();
}

int TqdcCoreConfig::getSimilarField(QMap<int, int> TqdcCoreConfig::*p) const
{
    QMap<int, int> f = this->*p;
    if(!f.isEmpty())
        return f.first();
    QMap<int, int> TqdcCoreConfig::* p1 = &TqdcCoreConfig::chDigThr;
    if(p == p1)
        return 0;
    return 0;
}

void TqdcCoreConfig::setGain(Gain value)
{
    for (auto &v : chGain)
        v = value;
}

void TqdcCoreConfig::makeTdcCoreConfig(TdcCoreConfig *tdc, DeviceId deviceId) const
{
    switch (deviceId) {
    case DEVICE_ID_TQDC16VS_E:;
        tdc->matchWin = tqdc_hptdc_scale*static_cast<quint16>(matchWin);
        tdc->latency = tqdc16vs_hptdc_lat_offset
            + tqdc_hptdc_scale*static_cast<quint16>(latency);
        break;
    default:
        tdc->matchWin = matchWin;
        tdc->latency = latency;
    }

    tdc->leadEdge = leadEdge;
    tdc->trailEdge = trailEdge;

    switch (deviceId) {
    case DEVICE_ID_TDC64VLE:
        tdc->adjustConfig.resolution = TDC_RES_INDEX_100;
        break;
    default:
        tdc->adjustConfig.resolution = TDC_RES_INDEX_25;
    }

    tdc->hptdcHdr = hptdcHdr;
    tdc->hptdcTrl = hptdcTrl;
    tdc->adjustConfig.enable_pair = false;
    tdc->chTdcEn = chTdcEn;
}

void TqdcCoreConfig::makeWaveRecConfig(WaveRecConfig *waveRec, DeviceId deviceId) const
{
    switch (deviceId) {
    case DEVICE_ID_TQDC16VS_E:
        waveRec->latency = tqdc16vs_adc_lat_offset + tqdc16vs_adc_scale * latency;
        waveRec->matchWin = tqdc16vs_adc_scale * matchWin;
        break;
    default:
        waveRec->matchWin = matchWin;
        waveRec->latency = latency;
    }
    waveRec->chAdcEn = chAdcEn;
    waveRec->chSignalInvert = chWfInvert;
    waveRec->chZsEn = chDigZsEn;
    waveRec->chZsRising = chDigZsRising;
    waveRec->chZsThr = chDigZsThr;
}

void TqdcCoreConfig::makeWaveTrigConfig(WaveTrigConfig *waveTrig) const
{
    QList<int> chList = chDigTrigEn.keys()+chDigThr.keys()+chDigTrigRising.keys();
    int maxCh = *std::max_element(chList.cbegin(), chList.cend());
    for(auto ch=0; ch<=maxCh; ++ch) {
        WaveTrigChannelConfig chCfg;
        chCfg.en = chDigTrigEn.value(ch, DEF_DIG_TRIG_EN);
        chCfg.edge = chDigTrigRising.value(ch, DEF_DIG_TRIG_RISING)
                ? WaveTrigChannelConfig::Edge::Rising
                : WaveTrigChannelConfig::Edge::Falling;
        chCfg.trigLevel = chDigThr.value(ch, DEF_DIG_THR);
        waveTrig->setChannel(ch, chCfg);
    }
}

void TqdcCoreConfig::makeBoardGpioConfig(BoardGpioConfig *boardGpio) const
{
    boardGpio->channels.clear();
    for (auto it = chGain.begin(); it != chGain.end(); ++it) {
        auto ch = it.key();
        auto &channel = boardGpio->channels[ch];
        channel.tri = false;
        channel.out = it.value() == Gain1;
    }
}

void TqdcCoreConfig::makeHitCoreConfig(HitCoreConfig *hitCore) const
{
    hitCore->polarity = 0;
    for (auto it = chInv.cbegin(); it != chInv.cend(); ++it) {
        quint64 ch = it.key();
        if (it.value())
            hitCore->polarity |= (1ull << ch);
    }
}
