#include "WrStatusModule.h"

#include "base/DeviceId.h"
#include "regio/regop.h"

#include "WrStatus.h"
#include "sdb/AfiSdbId.h"

// MCU SPI Slave
enum {
    REG_WR_EP_STATUS_CODE            = 0x00,
//    REG_WR_EP_STATUS_RESERVED        = 0x01,
    REG32_WR_EP_STATUS_UPTIME        = 0x02, // 32-bit
    REG64_WR_EP_STATUS_TIMECODE      = 0x04, // 64-bit
    REG32_WR_EP_STATUS_LINK_LOSS_CNT = 0x08, // 32-bit
    REG32_WR_EP_STATUS_SYNC_LOSS_CNT = 0x0A, // 32-bit
    REG32_WR_EP_STATUS_CODE_ERR_CNT  = 0x0C, // 32-bit
    REG32_WR_EP_STATUS_OTHER_ERR_CNT = 0x0E, // 32-bit

    REG64_WR_EP_SAVE_TIMECODE_BASE   = 0x10, // 64-bit
    REG64_WR_EP_SAVE_TIMECODE_CNT    = 4,
};

QList<AFI_SDB_ID> WrStatusModule::getSupportedSdbIds()
{
    return { AFI_SDB_WR_STATUS };
}

void WrStatusModule::post_init()
{
    if (!isAvailable()) {
        // no-SDB compatibility
        switch (getDeviceId()) {
        case DEVICE_ID_TDC64VHLE:
        case DEVICE_ID_TDC72VXS:
        case DEVICE_ID_TDC64VLE:
        case DEVICE_ID_TQDC16:
        case DEVICE_ID_TQDC16VS_E:
        case DEVICE_ID_TTVXS:
            bus.setAddrRange({0x0800, 0x001F});
            break;
        case DEVICE_ID_ADC64:
        case DEVICE_ID_ADC64ECAL:
        case DEVICE_ID_ADC64S2:
        case DEVICE_ID_ADC64VE:
        case DEVICE_ID_ADC64VE2:
        case DEVICE_ID_ADC64VE_XGE:
        case DEVICE_ID_ADC64WR:
        case DEVICE_ID_U40VE_RC:
        case DEVICE_ID_UT24VE:
            bus.setAddrRange({0x0060, 0x001F});
            break;
        default:;
        }
    }
}

mlink::RegOpVector WrStatusModule::getStatusRegOp(WrStatus *result) const
{
    Q_ASSERT(result);
    if (!result)
        return {};
    RegOpVector r;
    r.RegRead16(REG_WR_EP_STATUS_CODE, &result->statusCode.all);
    r.RegRead32(REG32_WR_EP_STATUS_UPTIME, &result->uptime);
    r.RegRead64(REG64_WR_EP_STATUS_TIMECODE, &result->timeCodeRaw);
    r.RegRead32(REG32_WR_EP_STATUS_LINK_LOSS_CNT, &result->linkError);
    r.RegRead32(REG32_WR_EP_STATUS_SYNC_LOSS_CNT, &result->syncFail);
    r.RegRead32(REG32_WR_EP_STATUS_CODE_ERR_CNT, &result->codeError);
    r.RegRead32(REG32_WR_EP_STATUS_OTHER_ERR_CNT, &result->otherError);
    for(uint16_t i=0; i<REG64_WR_EP_SAVE_TIMECODE_CNT; ++i)
        r.RegRead64(REG64_WR_EP_SAVE_TIMECODE_BASE + i*4, &result->savedTimeCodeRaw[i]);

    return r;
}

std::optional<WrStatus> WrStatusModule::readStatus()
{
    WrStatus status;
    if (!readStatus(&status))
        return {};
    return status;
}

bool WrStatusModule::readStatus(WrStatus *result)
{
    WrStatus status;
    auto regOp = getStatusRegOp(&status);
    auto t1 = QDateTime::currentMSecsSinceEpoch();
    const auto &opResult = regOpExecRebased(regOp);
    auto t2 = QDateTime::currentMSecsSinceEpoch();
    if (!opResult)
        return {};
    status.acquiredAtMs = (t1 + t2) /2;
    status.acquireErrorMs = std::abs(t1 - t2);
    if (result)
        *result = status;
    return true;
}
