#include "AdcClkStatus.h"

#include <QDebug>

static const int nx = 32;
static const int ny = 32;

namespace {
constexpr char symbol_opt[] = " █";
constexpr char symbol_ok[] = " ░";
constexpr char symbol_er[] = " ·";

static int vindex(int x, int y)
{
    return x + ny * y;
}

static void printMap(const QVector<bool> &map)
{
    QString s;
    for (int y=0; y<ny; y++) {
        s += QString("%1").arg(y, 2);
        for (int x=0; x<nx; x++) {
            s += QString(" %1").arg(map[vindex(x, y)]);
        }
        s += "\n";
    }
    qDebug().noquote() << s;
}

static QVector<bool> shrinkFilter(const QVector<bool> &v)
{
    QVector<bool> result(v.size(), false);
    for (int x=1; x<nx-1; x++)
        for (int y=1; y<ny-1; y++) {
            result[vindex(x, y)] = (
                v[vindex(x, y)] &
                v[vindex(x-1, y)] &
                v[vindex(x+1, y)] &
                v[vindex(x, y-1)] &
                v[vindex(x, y+1)]);
        }
    return result;
}

static QVector<bool> shrinkUntilEmpty(const QVector<bool> &data)
{
    QVector<bool> result(data);
    QVector<bool> map(data);
    while (true) {
        bool found = false;
        QVector<QPair<int, int> > v;
        for (int i=0; i<map.size(); i++)
            if (map[i]) {
                found = true;
                break;
            }
        if (!found)
            break;
        result = map;
        map = shrinkFilter(map);
    }
    return result;
}
} // namespace

AdcClkStatus::WPoint AdcClkStatus::getOptimalDelays(int adcId) const
{
    WPoint result {0, 0, 1e10, false};
    if (!data.contains(adcId))
        return result;
    //    QMap<int, QMap<int, bool > > data; // delayStatus <clockDelay, <dataDelay, isOk> >
    const auto &adcIdData = data[adcId];
    QVector<bool> map(nx*ny);
    for (int x=0; x<nx; x++)
        for (int y=0; y<ny; y++)
            map[vindex(x, y)] = adcIdData[y][x];
    map = shrinkUntilEmpty(map);

    // printMap(map);
    QVector<WPoint> wpoints;
    for (int i=0; i<map.size(); i++)
        if (map[i]) {
            int x = i % ny;
            int y = i / ny;
            // more weight to delay delta, less to delays sum
            wpoints.push_back({x, y, 1./4. * abs(x-y) + 1./64. * (x+y), true});
        }
    for (const auto &i : wpoints)
        if (i.weight < result.weight)
            result = i;
    return result;
}

AdcClkStatus AdcClkStatus::aggregate(QList<AdcClkStatus> statusList)
{
    AdcClkStatus res;

    for(auto &r : statusList) {
        if(res.data.isEmpty()) {
            res = r;
            continue;
        }
        for (int adcId : r.data.keys()) {
            for (int clockDelay : r.data[adcId].keys()) {
                for (int dataDelay : r.data[adcId][clockDelay].keys()) {
                    res.data[adcId][clockDelay][dataDelay] &= r.data[adcId][clockDelay][dataDelay];
                }
            }
        }
    }
    return res;
}

QString AdcClkStatus::toString() const
{
    QString s;
    for (int adcId : data.keys()) {
        const auto &optPoint = getOptimalDelays(adcId);
        s += QString("[AdcId: %1] ").arg(adcId);
        if(optPoint.status) {
            s += QString("Optimal delays: clockDelay(%1) \\ dataDelay(%2)")
                    .arg(optPoint.y).arg(optPoint.x);
        } else {
            s += "Optimal delays wasn't found";
        }
        s+= "\n";

        QString headerStr = "    ";
        for(int i=0; i<nx; ++i)
            headerStr += QString("%1").arg(i%10, 2);
        s += headerStr + "\n";
        for (int clockDelay : data[adcId].keys()) {
            QString devStatusStr = QString("%1: ").arg(clockDelay, 2);
            for (int dataDelay : data[adcId][clockDelay].keys()) {
                bool opt = (optPoint.y == clockDelay && optPoint.x == dataDelay);
                bool good = data[adcId][clockDelay][dataDelay];
                devStatusStr.append(good ? (opt ? symbol_opt : symbol_ok) : symbol_er);
            }
            s += devStatusStr + "\n";
        }
    }
    return s;
}

QString AdcClkStatus::getLegend()
{
    QStringList l;
    l << QString("%1 - %2").arg(symbol_opt).arg("optimal delays");
    l << QString("%1 - %2").arg(symbol_ok).arg("good delays");
    l << QString("%1 - %2").arg(symbol_er).arg("bad delays");
    return l.join("\n");
}

AdcSerDesDelayMap AdcClkStatus::getDevDelay() const
{
    AdcSerDesDelayMap dd;
    for(auto adcId : data.keys()) {
        if(adcId<0)
            continue; // skip agregated data
        const auto &wPoint = getOptimalDelays(adcId);
        if(wPoint.status)
            dd.map[adcId] = {wPoint.y, wPoint.x};
    }
    return dd;
}
