//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Client.h"

#include "base/DnsLookup.h"
#include "base/RunDocument.h"
#include "base/RunDocumentConverter.h"
#include "remote-control-server/RcProgramState.h"
#include "remote-control-server/RemoteControlProtocol.h"
#include "remote-control-server/RunControlCommands.h"

#include <QTcpSocket>
#include <QJsonDocument>
#include <QJsonObject>
#include <QDebug>
#include <QLoggingCategory>
#include <QTimer>

#if QT_VERSION < 0x050000
Q_DECLARE_METATYPE(QAbstractSocket::SocketError)
#endif

namespace {
Q_DECLARE_LOGGING_CATEGORY(cli_mgr)
Q_LOGGING_CATEGORY(cli_mgr, "afi.rc.cli_mgr")

const qint32 EXPIRE_TIMEOUT = 1000;
}

Client::Client(const ClientIndex &index) :
    QObject(nullptr),
    index(index),
    enabled(false),
    tcpSocket(new QTcpSocket(this)),
    hostPort(0),
    reconnectTimer(new QTimer(this)),
    dnsLookup(new DnsLookup(this)),
    connectingAttempts(0),
    run(false),
    block_size(0),
    dataTimer(new QTimer(this))
{
    Q_ASSERT(!index.isEmpty());
    AbstractConfigConverter<RunDocument>::registerTypes();
    AbstractConfigConverter<RcProgramState>::registerTypes();
    qRegisterMetaType<QHostAddress>();
    qRegisterMetaType<RC_CLIENT_PROTO_ID>();
    qRegisterMetaType<RcClientStatus>();

    connect(tcpSocket, &QTcpSocket::readyRead,
            this, &Client::readCommandResponce, Qt::QueuedConnection);
    connect(tcpSocket, &QTcpSocket::disconnected,
            this, &Client::disconnected, Qt::QueuedConnection);
    qRegisterMetaType<QAbstractSocket::SocketError>();
    connect(tcpSocket, SIGNAL(error(QAbstractSocket::SocketError)),
            SLOT(displayError(QAbstractSocket::SocketError)), Qt::QueuedConnection);
    connect(reconnectTimer, &QTimer::timeout,
            this, &Client::reconnectToServer, Qt::QueuedConnection);
    reconnectTimer->setSingleShot(true);

    connect(dataTimer, &QTimer::timeout,
            this, &Client::checkData);
    dataTimer->setSingleShot(true);
    dataTimer->setInterval(EXPIRE_TIMEOUT);

    connect(dnsLookup, &DnsLookup::reverseEntryUpdated,
            this, &Client::reverseEntryUpdated);
}

void Client::readCommandResponce()
{
    while (tcpSocket->bytesAvailable()) {
        const auto d = tcpSocket->readAll();
        recivedData += d;
    }

    processIncomingMsg();
}

void Client::reconnectToServer()
{
    if(!enabled) return;

    try{
        if(tcpSocket->isValid() &&
                (tcpSocket->state() == QTcpSocket::ConnectedState ||
                 tcpSocket->state() == QTcpSocket::ConnectingState) ) {
            return;
        }
        if(tcpSocket->state() != QTcpSocket::UnconnectedState)
            tcpSocket->abort();
        if(hostAddr.isNull() || hostPort==0) {
            reconnectTimer->start(5000);
            return;
        }

        tcpSocket->connectToHost(hostAddr, hostPort);
        ++connectingAttempts;
        if(tcpSocket->waitForConnected(1000)){
            updateHostInStatus();
            qDebug() << getIdentStr() << "TcpClient connected";
            emit statusUpdated(index, "Connected");
        } else {
            if(connectingAttempts<=1) {
                QString mess = QString("TcpClient can't connected.Err:%2")
                        .arg(tcpSocket->errorString());
                qWarning() << getIdentStr() << mess;
//                sendLog(mess);
                emit statusUpdated(index, mess);
                reconnectTimer->start(1000);
            } else
                reconnectTimer->start(5000);
        }
    } catch(...) {
        qWarning() << getIdentStr() << "Exception in TcpClient::requestNewData()";
        reconnectTimer->start(5000);
    }
}

void Client::displayError(QAbstractSocket::SocketError socketError)
{
    switch (socketError) {
    case QAbstractSocket::RemoteHostClosedError:
        qInfo() << getIdentStr() << "Remote hoste closed connection.";
        emit statusUpdated(index, "Remote hoste closed connection");
        break;
    case QAbstractSocket::HostNotFoundError:
        qInfo() << getIdentStr() << "The host was not found."
                << "Please check the host name and port settings.";
        emit statusUpdated(index, "host was not found");
        break;
    case QAbstractSocket::ConnectionRefusedError:
        if(connectingAttempts<=1){
            qInfo() << getIdentStr()
                    << tr("The connection was refused by the peer. "
                   "Make sure the fortune server is running, "
                   "and check that the host name and port "
                   "settings are correct.");
            emit statusUpdated(index, "The connection was refused by the peer");
        }
        break;
    default:
        qInfo() << getIdentStr()
                << tr("The following error occurred: %1.")
                .arg(tcpSocket->errorString());
        emit statusUpdated(index, tr("The following error occurred: %1.")
                        .arg(tcpSocket->errorString()));
    }
}

void Client::sendRunState(FsmState state, const RcRunId &_runId, const QVariant &evNum)
{
    if(!tcpSocket->isValid() ||
        (tcpSocket->state() != QTcpSocket::ConnectedState &&
         tcpSocket->state() != QTcpSocket::ConnectingState) ){
        emit fsmStateUpdated(index, FsmState::Unknown);
        return;
    }

    switch (state) {
    case FsmState::Starting:
        run = true;
        runId = _runId;
        break;
    case FsmState::Stopping:
        run = false;
        break;
    default:
        qCritical() << getIdentStr() << __func__ << "Invalid FsmState:" << state;
        return;
    }

    if(protocolId == RC_CLIENT_PROTO_V5) {
        QJsonObject obj;
        obj[RC_JSON_RUN_ID] = QVariant::fromValue(runId).value<QJsonObject>();
        if(evNum.isValid()) {
            obj[RC_JSON_RUN_EV_NUM] = jsonValueFromUInt64(evNum.toULongLong());
        }
        send_rc_json(tcpSocket, {{run ? RC_CMD_START : RC_CMD_STOP, obj}});
    } else {
        sendCommand({run ? RC_CMD_START : RC_CMD_STOP,
                     runId.runIndex, QString::number(runId.runNumber),
                     runId.uuid.toString()});
    }
}

void Client::startReadOut(const ClientIndex &i)
{
    if(!i.isEmpty() && index != i) {
        return;
    }

    if(enabled) {
        qWarning() << getIdentStr() << "Trying to start started client";
        return;
    }
    reconnectToServer();
    enabled = true;
    reconnectTimer->start(0);
    emit statusUpdated(index, "connecting to server...");
}


void Client::stopReadOut(const ClientIndex &i)
{
    if(!i.isEmpty() && index != i) {
        return;
    }
    if(!enabled) {
        qWarning() << getIdentStr() <<"Trying to stop stopped client";
        return;
    }
    enabled = false;
    protocolId = RC_CLIENT_PROTO_UNKNOWN;
    emit protocolChanged(index, protocolId);
    if(tcpSocket->state() != QTcpSocket::UnconnectedState)
        tcpSocket->disconnectFromHost();
    reconnectTimer->stop();
    emit statusUpdated(index, "tcp: stoped readout");
}

void Client::setEvNumCheck(const ClientIndex &i, bool evCheck)
{
    if(index != i)
        return;
    performEvCheck = evCheck;
}

void Client::deleteClient(const ClientIndex &i)
{
    if(!i.isEmpty() && index != i)
        return;
    deleteLater();
}

void Client::setAddress(const ClientIndex &i, const QHostAddress &addr, quint16 port)
{
    if(index != i)
        return;
    if(hostAddr == addr && hostPort == port)
        return;
    // Different addr/port can send program from other host
    if(isConnectionAlive())
        return;
    bool enabledBak = enabled;
    if(enabled){
        blockSignals(true);
        stopReadOut(index);
        blockSignals(false);
    }
    hostAddr = addr;
    hostPort = port;

    if(enabledBak)
        startReadOut(index);
}

void Client::setEmptyAddress(const ClientIndex &i)
{
    if(index != i)
        return;
    hostAddr.clear();
    hostPort = 0;
    updateHostInStatus();
}

void Client::evNumRequest(int checkIndex)
{
    if(!enabled || !performEvCheck)
        return;
    if(!isConnectionAlive()) {
        emit fsmStateUpdated(index, FsmState::Unknown);
        return;
    }
    switch (protocolId) {
    case RC_CLIENT_PROTO_V5:
    {
        QJsonObject obj;
        obj.insert(RC_JSON_EV_NUM_CHECK_ID, checkIndex);
        send_rc_json(tcpSocket, {{RC_CMD_EV_NUM_CHECK, obj}});
        break;
    }
    default:
        sendCommand({RC_CMD_EV_NUM_CHECK, QString::number(checkIndex)});
        break;
    }
}

void Client::sendRunDocumentToClients(const RunDocument &rd)
{
    if(!enabled)
        return;
    static const QList<DaqModule::Type> allowedProgs{
        DaqModule::Type::RunControl,
        DaqModule::Type::Ttvxs,
        DaqModule::Type::EventBuilder,
    };
    if(!allowedProgs.contains(DaqModule::Type(index.program_type)))
        return;

    if(!isConnectionAlive()) {
        emit fsmStateUpdated(index, FsmState::Unknown);
        return;
    }
    switch (protocolId) {
    case RC_CLIENT_PROTO_V1:
    case RC_CLIENT_PROTO_V2:
    case RC_CLIENT_PROTO_V3:
        qWarning() << getIdentStr() << "Old protocol are unsupported";
        break;
    case RC_CLIENT_PROTO_V4:{
        auto json = QVariant::fromValue(rd).value<QJsonObject>();
        QJsonObject obj{{RC_JSON_RUN_DOC, json}};
        const auto &data = QJsonDocument(obj).toJson(QJsonDocument::Compact);
        send_run_document(tcpSocket, data);
        break;
    }
    case RC_CLIENT_PROTO_V5:{
        auto rdObj = QVariant::fromValue(rd).value<QJsonObject>();
        QJsonObject obj;
        obj.insert(RC_JSON_VAL, rdObj);
        send_rc_json(tcpSocket, {{RC_JSON_RUN_DOC, obj}});
        break;
    }
    default:
        qWarning() << getIdentStr() << "Can't send config in case of FORMAT_ID_UNKNOWN";
    }
}

void Client::sendRcProgramState(const RcProgramState &allState) const
{
    if(!enabled)
        return;

    if(!isConnectionAlive()) {
        emit fsmStateUpdated(index, FsmState::Unknown);
        return;
    }
    switch (protocolId) {
    case RC_CLIENT_PROTO_V1:
    case RC_CLIENT_PROTO_V2:
    case RC_CLIENT_PROTO_V3:
    case RC_CLIENT_PROTO_V4:
        qWarning() << getIdentStr() << "Old protocol are unsupported";
        break;
    case RC_CLIENT_PROTO_V5:
    {
        send_rc_json(tcpSocket, {{RC_CMD_TREE_STATE, QVariant::fromValue(allState).toJsonValue()}});
        break;
    }
    default:
        qWarning() << getIdentStr() << "Can't send config in case of FORMAT_ID_UNKNOWN";
    }
}

void Client::disconnected()
{
    recivedData.clear();
    if(enabled) {
        QString s = "TcpClient was disconnected, trying to reconnect";
        (run?qWarning():qInfo()) << getIdentStr() << s;
        emit statusUpdated(index, s);
        emit fsmStateUpdated(index, FsmState::Unknown);
        if(run)
            emit runError(index, runId, QString("Client was disconnected"));
        connectingAttempts=0;
        reconnectTimer->start(1000);
    } else {
        qInfo() << getIdentStr() << "TcpClient disconnected.";
    }
}

void Client::checkData()
{
    if(!recivedData.isEmpty()) {
        qWarning() << getIdentStr() << "has timeout on incompleted data" << "recivedData.size()="<<recivedData.size();
        dataTimerTimeout=true;
        recivedData.remove(0,1);
        ++curruptedSize;
        processIncomingMsg();
    }
    qWarning() << getIdentStr() << "curruptedSize " << curruptedSize;
}

void Client::reverseEntryUpdated(const DnsReverseEntry &ent)
{
    bool ok;
    const auto ip = status.host.toIPv4Address(&ok);
    Q_ASSERT(ok);
    if(ent.first == ip && status.hostName != ent.second) {
        status.hostName = ent.second;
        emit statusChanged(index, status);
    }
}

void Client::sendCommand(const QStringList &commandList)
{
    switch (protocolId) {
    case RC_CLIENT_PROTO_V1:
    case RC_CLIENT_PROTO_V2:
    case RC_CLIENT_PROTO_V3:
        qWarning() << getIdentStr() << "Old protocol are unsupported";
        break;
    case RC_CLIENT_PROTO_V4:
        send_rc_command(tcpSocket, commandList);
        break;
    default:
        qWarning() << getIdentStr() << "Can't send command in case of FORMAT_ID_UNKNOWN";
    }
}

bool Client::isConnectionAlive() const
{
    if(tcpSocket->isValid() &&
            (tcpSocket->state() == QTcpSocket::ConnectedState ||
             tcpSocket->state() == QTcpSocket::ConnectingState) )
        return true;
    return false;
}

void Client::updateHostInStatus()
{
    auto h = isConnectionAlive() ? hostAddr : QHostAddress();
    if(status.host == h)
        return;
    status.host = h;
    status.hostName.clear();
    dnsLookup->reverseLookup(status.host.toString());
    emit statusChanged(index, status);
}

bool Client::processIncomingMsg()
{
    while (!recivedData.isEmpty()) {
        bool ok = true;
        if(protocolId == RC_CLIENT_PROTO_V4 || protocolId == RC_CLIENT_PROTO_V5 || protocolId == RC_CLIENT_PROTO_UNKNOWN) {
            ok = processIncomingMsgV4_5();
        }

        if(ok) {
            dataTimer->stop();
            dataTimerTimeout = false;
        } else {
            if(dataTimerTimeout) {
                if(protocolId != RC_CLIENT_PROTO_UNKNOWN) {
                    protocolId = RC_CLIENT_PROTO_UNKNOWN;
    //                qWarning() << getIdentStr() <<  "communication sync. fail or big incoming data block";
                    emit protocolChanged(index, protocolId);
                    continue; // try again both of them
                }
                recivedData.remove(0,1);
                ++curruptedSize;
            } else {
                break;
            }
        }
    }

    if(recivedData.isEmpty()) {
        dataTimer->stop();
        dataTimerTimeout = false;
    } else {
        if(!dataTimer->isActive()) {
            dataTimer->start();
        }
    }
    return true;
}

bool Client::processIncomingMsgV4_5()
{
    if(recivedData.size() < static_cast<int>(sizeof(TypeLengthHeader)))
        return false;

    auto hdr = reinterpret_cast<const TypeLengthHeader *>(recivedData.data());
    const auto &type = hdr->type;

    bool ok = (type == RC_MSG_CONFIG_V4) || (type == RC_MSG_CMD_V4)
            || (type == RC_MSG_JSON_V5);
    if(!ok)
        return ok;

    if(hdr->len+sizeof (TypeLengthHeader) > static_cast<quint32>(recivedData.size())) {
        if(!recivedData.isEmpty()) {
            if(!dataTimer->isActive()) {
                dataTimer->start();
            }
        }
        return false;
    }

    const auto &data_ba = recivedData.mid(sizeof (TypeLengthHeader),
                                         static_cast<int>(hdr->len));
    switch (type) {
    case RC_MSG_CONFIG_V4:
        //        qDebug(cli_mgr) << __PRETTY_FUNCTION__ << getIdentStr() << hex << type << dec << "cfgSize:" << data_ba.size();
        config_parse(data_ba);
        break;
    case RC_MSG_CMD_V4:
        //        qDebug(cli_mgr) << __PRETTY_FUNCTION__ << getIdentStr() << hex << type << dec << "data" << QString::fromUtf8(data_ba);
        command_line_parse(data_ba);
        break;
    case RC_MSG_JSON_V5:
        //        qDebug(cli_mgr) << __PRETTY_FUNCTION__ << getIdentStr() << hex << type << dec << "data" << QString::fromUtf8(data_ba);
        command_json_parse(data_ba);
        break;
    default:;
    }

    if(ok) {
        auto blockProtocol = (type == RC_MSG_JSON_V5) ? RC_CLIENT_PROTO_V5 : RC_CLIENT_PROTO_V4;
        recivedData.remove(0, static_cast<int>(hdr->len+sizeof (TypeLengthHeader)));
        if(protocolId != blockProtocol) {
            protocolId = blockProtocol;
            qInfo() << getIdentStr() << "set communication protocol"
                    << getProtocolIdStr(protocolId);
            emit protocolChanged(index, protocolId);
        }
    }

    return ok;
}

bool Client::command_line_parse(QByteArray data)
{
    bool allOk=true;
    QString message = QString::fromUtf8(data);

    QStringList lines = message.split(RC_CMD_END, QString::SkipEmptyParts);
    if(lines.isEmpty())
        return false;
    for(QString line: lines) {
        bool lineParsed=true;
        line = line.trimmed();
        QStringList lineParts = line.split(" ", QString::SkipEmptyParts);
        if(lineParts.isEmpty())
            continue;
//        qDebug(cli_mgr) << getIdentStr() <<"got from client: "<<line;
        if(lineParts[0] == RC_CMD_STATE || lineParts[0] == RC_CMD_STATUS) {
            QString statusStr = line.right(line.size() - QString(RC_CMD_STATE).size()).trimmed();
            emit statusUpdated(index, statusStr);
        } else if(lineParts[0] == RC_CMD_FSM_STATE) {
            const auto &stateStr = line.right(line.size() - QString(RC_CMD_FSM_STATE).size()).trimmed();
            fsmState = getFsmState(stateStr);
            emit fsmStateUpdated(index, fsmState);
        } else if(lineParts[0] == RC_CMD_EV_NUM_CHECK) {
            lineParsed = lineParts.size()>1;
            int checkIndex = 0;
            EvNumSet evNums;
            quint64 evNum;
            if(lineParsed)
                checkIndex = lineParts[1].toInt(&lineParsed);
            if(lineParsed) {
                for(int i=2; i<lineParts.size(); ++i){
                    evNum = lineParts[i].toULongLong(&lineParsed);
                    if(!lineParsed)
                        break;
                    evNums.insert(evNum);
                }
            }
            if(lineParsed)
                emit evNumResponse(checkIndex, index, evNums);
        } else if(lineParts[0] == RC_CMD_ERROR) {
            lineParsed = (lineParts.size()>=3);
            int runNum = 0;
            if(lineParsed)
                runNum = lineParts[2].toInt(&lineParsed);
            if(lineParsed) {
                QString errStr = line;
                for(int i=0;i<3;++i)
                    errStr.replace(errStr.indexOf(lineParts[i]), lineParts[i].size(),"");
                emit runError(index, RcRunId(lineParts[1], runNum), errStr.trimmed());
            }
        } else if(lineParts[0] == RC_CMD_WARN) {
            lineParsed = (lineParts.size()>=3);
            int runNum = 0;
            if(lineParsed)
                runNum = lineParts[2].toInt(&lineParsed);
            if(lineParsed) {
                QString warnStr = line;
                for(int i=0;i<3;++i)
                    warnStr.replace(warnStr.indexOf(lineParts[i]), lineParts[i].size(),"");
                emit runWarning(index, RcRunId(lineParts[1], runNum), warnStr.trimmed());
            }
        } else if(lineParts[0] == RC_CMD_BACK_PRESSURE) {
            lineParsed = (lineParts.size()==4);
            int runNum = 0;
            int boolVal;
            if(lineParsed)
                runNum = lineParts[2].toInt(&lineParsed);
            if(lineParsed) {
                boolVal = lineParts[3].toInt(&lineParsed);
                emit backPressure(index, RcRunId(lineParts[1], runNum), boolVal);
            }
        } else if(lineParts[0] == RC_CMD_VME_FLAG) {
            lineParsed = (lineParts.size()==2);
            int hasVme;
            if(lineParsed) {
                hasVme = lineParts[1].toInt(&lineParsed);
                emit vmeFlag(index, hasVme);
            }
//        } else {
//            // Other possible command is success/fail
//            bool success = false;
//            RUN_STATES rs = RS_ERROR;
//            QStringList lineParts = line.split(" ", QString::SkipEmptyParts);
//            lineParsed = (lineParts.size()==4);

//            if(lineParsed){
//                if(lineParts[0] == RC_CMD_SUCCESS_PREFIX) {
//                    success=true;
//                } else if(lineParts[0] == RC_CMD_FAILED_PREFIX) {
//                    success=false;
//                } else {
//                    // unknown command; skipp it
//                    lineParsed=false;
//                }
//            }

//            if(lineParsed) {
//                QString clRunIndex = lineParts[2];
//                int clRunNumber = QString(lineParts[3]).toInt(&lineParsed);

//                if(lineParts[1] == RC_CMD_START)
//                    rs = RS_STARTED;
//                else if(lineParts[1] == RC_CMD_STOP)
//                    rs = RS_STOPPED;
//                else
//                    lineParsed = false;

//                if(lineParsed)
//                    emit runStateUpdated(id, success ? rs : RS_ERROR,
//                                         RcRunId(clRunIndex, clRunNumber));
//            }
        }
        if(!lineParsed){
            allOk = false;
            if(protocolId != RC_CLIENT_PROTO_UNKNOWN)
                qWarning() << "Can't parse line from client "
                           << getIdentStr() <<": "<<line;
        }
    }
    return allOk;
}

void Client::config_parse(QByteArray data)
{
    const auto &json = QJsonDocument::fromJson(data).object();
    qInfo() << QString("%1 got JSON config (size=%L2)")
               .arg(getIdentStr()).arg(data.size());
    emit config_transfer(index, json);
}

bool Client::command_json_parse(const QByteArray &data)
{
    QJsonParseError error;
    auto topObj = QJsonDocument::fromJson(data, &error).object();
    if(error.error != QJsonParseError::NoError) {
        qWarning().noquote()
                << QString("Error while convert incoming data to JSON")
                   .arg(RC_MSG_CONFIG_V4, 8, 16, QChar('0'));
        return false;
    }

    bool allOk=true;

    qDebug(cli_mgr) << getIdentStr() << "got from client JSON: "
                    << QString::fromUtf8(data);

    for(auto it=topObj.constBegin(); it!=topObj.constEnd(); ++it) {
        const auto &key = it.key();

        if(key == RC_CMD_STATUS) {
            QString statusStr = it->toString();
            emit statusUpdated(index, statusStr);
            continue;
        }

        if(key == RC_CMD_FSM_STATE) {
            const auto &stateStr = it->toString();
            fsmState = getFsmState(stateStr);
            emit fsmStateUpdated(index, fsmState);
            continue;
        }
        if(key == RC_CMD_TREE_STATE) {
            const auto state = QVariant::fromValue(*it).value<RcProgramState>();
            emit rcProgramStateUpdated(index, state);
            continue;
        }

        if(key == RC_CMD_EV_NUM_CHECK) {
            const auto &jsObj = it->toObject();
            int checkIndex = jsObj[RC_JSON_EV_NUM_CHECK_ID].toInt();
            auto evNumArr = jsObj[RC_JSON_VAL].toArray();
            EvNumSet evNums;
            for(auto v : evNumArr) {
                evNums << jsonValueToUInt64(v);
            }
            emit evNumResponse(checkIndex, index, evNums);
            continue;
        }

        if(key == RC_CMD_ERROR) {
            const auto &jsObj = it->toObject();
            auto runId = QVariant::fromValue(jsObj[RC_JSON_RUN_ID]).value<RcRunId>();
            auto val = jsObj[RC_JSON_VAL].toString();
            emit runError(index, runId, val);
            continue;
        }

        if(key == RC_CMD_WARN) {
            const auto &jsObj = it->toObject();
            auto runId = QVariant::fromValue(jsObj[RC_JSON_RUN_ID]).value<RcRunId>();
            auto val = jsObj[RC_JSON_VAL].toString();
            emit runWarning(index, runId, val);
            continue;
        }

        if(key == RC_CMD_BACK_PRESSURE) {
            const auto &jsObj = it->toObject();
            auto runId = QVariant::fromValue(jsObj[RC_JSON_RUN_ID]).value<RcRunId>();
            auto val = jsObj[RC_JSON_VAL].toBool();
            emit backPressure(index, runId, val);
            continue;
        }

        if(key == RC_CMD_VME_FLAG) {
            const auto &val = it->toBool();
            emit vmeFlag(index, val);
            continue;
        }

        if(key == RC_JSON_RUN_DOC) {
            const auto &val = it->toObject()[RC_JSON_VAL].toObject();
            qInfo() << QString("%1 got JSON config (size=%L2)")
                       .arg(getIdentStr()).arg(data.size());
            emit config_transfer(index, val);
            continue;
        }

        if(unknownCommands.contains(key)) {
            qWarning() << QString("Got unknown command '%1'").arg(key) << *it;
            unknownCommands << key;
        }
    }

    return allOk;
}
