#include "RcClientTreeModel.h"

#include <QColor>
#include <QFont>

#include "RcProgramStateTreeItem.h"

namespace {
bool needReset(const RcProgramState &st2, const RcProgramState &st1)
{
    if(st1.index != st2.index)
        return true;
    auto it2=st2.clients.cbegin();
    for(auto it1=st1.clients.cbegin(); it1!=st1.clients.cend(); ++it1) {
        if(it2 ==st2.clients.cend())
            return true;
        if(it1.key() != it2.key())
            return true;
        bool clReset = needReset(*it1, *it2);
        if(clReset)
            return true;
        ++it2;
    }
    return it2!=st2.clients.cend();
}

RcProgramStateTreeItem *createNode(const RcProgramState &state)
{
    auto node = new RcProgramStateTreeItem;
    node->index = state.index;
    for(auto it=state.clients.cbegin();it!=state.clients.cend();++it) {
        auto ch = createNode(*it);
        ch->parent = node;
        node->children << ch;
    }
    return node;
}
}

RcClientTreeModel::RcClientTreeModel(QObject *parent)
    : QAbstractItemModel{parent}
    , mRootItem(new RcProgramStateTreeItem)
{}

RcClientTreeModel::~RcClientTreeModel()
{
    delete mRootItem;
}

QModelIndex RcClientTreeModel::index(int row, int column, const QModelIndex &parent) const
{
    if(!hasIndex(row, column, parent))
        return {};
    auto parentItem = parent.isValid() ?
        static_cast<RcProgramStateTreeItem*>(parent.internalPointer()) :
        mRootItem;
    auto childItem = parentItem->child(row);
    if(childItem) {
        return createIndex(row, column, childItem);
    }
    return {};
}

QModelIndex RcClientTreeModel::parent(const QModelIndex &child) const
{
    if (!child.isValid())
        return QModelIndex();

    auto childItem = static_cast<RcProgramStateTreeItem*>(child.internalPointer());
    auto parentItem = childItem->parent;

    if (parentItem == mRootItem)
        return QModelIndex();

    if(parentItem)
        return createIndex(parentItem->row(), 0, parentItem);
    return {};
}

int RcClientTreeModel::rowCount(const QModelIndex &parent) const
{
    auto parentItem = parent.isValid() ?
        static_cast<RcProgramStateTreeItem*>(parent.internalPointer()) :
        mRootItem;

    return parentItem->childCount();
}

int RcClientTreeModel::columnCount(const QModelIndex &parent) const
{
    return COL_TOTAL;
}

QVariant RcClientTreeModel::data(const QModelIndex &index, int role) const
{
    if (!index.isValid())
        return QVariant();

    auto item = static_cast<RcProgramStateTreeItem*>(index.internalPointer());

    switch(role) {
    case Qt::DisplayRole: return data_DisplayRole(item, index.column());
    case Qt::CheckStateRole: return data_CheckStateRole(item, index.column());
    case Qt::ForegroundRole: return data_ForegroundRole(item, index.column());
    case Qt::FontRole: return data_FontRole(item, index.column());

    default:;
    }

    return {};
}

bool RcClientTreeModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    if (!index.isValid())
        return false;

    switch(role) {
    case Qt::CheckStateRole:
    {
        auto item = static_cast<RcProgramStateTreeItem*>(index.internalPointer());
        switch (value.value<Qt::CheckState>()) {
        case Qt::Checked:
            globTrigs << item->index;
            return true;
        case Qt::Unchecked:
            if(globTrigs.remove(item->index))
                emit globTrigsChanged(globTrigs);
            return true;
        default:;
        };
        return false;
    }
    default:;
    }

    return {};
}

Qt::ItemFlags RcClientTreeModel::flags(const QModelIndex &index) const
{
    if(!index.isValid())
        return Qt::NoItemFlags;
    auto f = QAbstractItemModel::flags(index);

    switch(index.column()) {
    case COL_IS_GLOB_TRIG:
        f |= Qt::ItemIsUserCheckable | Qt::ItemIsEnabled;
        break;
    default:;
    }

    if(!enable)
        f = f & (~Qt::ItemIsEnabled);

    return f;
}

QVariant RcClientTreeModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
        switch (section) {
        case COL_NAME: return "Name";
        case COL_FSM: return "FSM";
        case COL_IS_GLOB_TRIG: return "GlobTrig";
        }
    }
    return {};
}

void RcClientTreeModel::setEnabled(bool en)
{
    if(enable == en)
        return;

    enable = en;
    auto top = index(0,0);
    auto lastDev = index(rowCount(top)-1, columnCount(top)-1, top);
    auto lastCh = index(rowCount(lastDev)-1, columnCount(lastDev)-1, lastDev);
    emit dataChanged(top, lastCh);
}

void RcClientTreeModel::clientRcProgramStateUpdated(const RcProgramState &state)
{
    if(rcProgramState == state)
        return;

    bool doReset = needReset(rcProgramState, state);
    if(doReset)
        beginResetModel();

    rcProgramState = state;
    if(doReset) {
        qDeleteAll(mRootItem->children);
        if(!mRootItem->children.isEmpty())
            mRootItem->children.clear();

        auto topNode = createNode(rcProgramState);
        mRootItem->children << topNode;
        endResetModel();
    } else
        sendDataChanged();
}

void RcClientTreeModel::setMyClientIndex(const ClientIndex &index)
{
    if(myClientIndex == index)
        return;
    ClientIndexSet cis{myClientIndex, index};
    const auto &mis = clientIndexToModelIndex(&cis);
    myClientIndex = index;
    for(auto i : mis)
        emit dataChanged(i,i, {});
}

const RcProgramState *RcClientTreeModel::getStateByPath(const RcClientPath &p) const
{
    const RcProgramState* ptr = &rcProgramState;
    for(auto it=p.cbegin(); it!=p.cend(); ++it) {
        if(it==p.cbegin()) {
            continue;
        }
        auto key = *it;
        auto &map = ptr->clients;
        auto pIt = map.constFind(key);
        if(pIt == map.cend())
            return nullptr;
        ptr = &(*pIt);
    }
    return ptr;
}

QVariant RcClientTreeModel::data_DisplayRole(const RcProgramStateTreeItem *item, int column) const
{
    switch(column) {
    case COL_NAME:
        return QString("%1 #%2")
            .arg(item->index.getProgramType().toString())
            .arg(item->index.getProgramIndex().toString());
        return item->index.toString();
    case COL_FSM:
    {
        auto ptr = getStateByPath(item->getPath());
        if(ptr)
            return fsmStateToString(ptr->fsm);
        return {};
    }
    default: return {};
    }
}

QVariant RcClientTreeModel::data_CheckStateRole(const RcProgramStateTreeItem *item, int column) const
{
    switch(column) {
    case COL_IS_GLOB_TRIG:
    {
        return globTrigs.contains(item->index) ? Qt::Checked : Qt::Unchecked;
    }
    default: return {};
    }
}

QVariant RcClientTreeModel::data_ForegroundRole(const RcProgramStateTreeItem *item, int column) const
{
    switch(column) {
    case COL_NAME:
    case COL_FSM:
        return itemColor(item);
    default:;
    }
    return {};
}

QVariant RcClientTreeModel::data_FontRole(const RcProgramStateTreeItem *item, int column) const
{
    switch(column) {
    case COL_NAME:
        if(item->index == myClientIndex) {
            QFont font;
            font.setBold(true);
            return font;
        }
        break;
    default:;
    }
    return {};
}

QVariant RcClientTreeModel::itemColor(const RcProgramStateTreeItem *item) const
{
    if(auto ptr = getStateByPath(item->getPath())) {
        switch(ptr->fsm) {
        case FsmState::Ready:
        case FsmState::Run:
            return QColor(Qt::darkGreen);
            return QColor(60, 179, 113);
        case FsmState::Error:
        case FsmState::Unknown:
            return QColor(Qt::red);
            return QColor(255, 99, 71);
        default:;
        }
    }

    return {};
}

void RcClientTreeModel::sendDataChanged(const QModelIndex &i)
{
    auto rows = i.row();
    if(rows == 0)
        return;

    auto top = index(0, 0, i);
    auto last = index(rows-1, columnCount(i)-1, i);
    emit dataChanged(top, last);

    for(auto r=0; r<i.row(); ++r) {
        sendDataChanged(index(r,0, i));
    }
}

QSet<QModelIndex> RcClientTreeModel::clientIndexToModelIndex(ClientIndexSet *cis, const QModelIndex &i) const
{
    QSet<QModelIndex> mis;
    if(i.isValid()) {
        auto ii = static_cast<RcProgramStateTreeItem*>(i.internalPointer())->index;
        if(cis->remove(ii)) {
            mis << i;
            if(cis->isEmpty())
                return mis;
        }
    }
    if(auto rows = rowCount(i)) {
        for(auto r=0; r<rows; ++r) {
            for(auto mi : clientIndexToModelIndex(cis, index(r, 0, i)))
                mis.insert(mi);
        }
    }
    return mis;
}
