#include "BaseConfig.h"

#include <QCoreApplication>
#include <QDateTime>
#include <QDebug>
#include <QElapsedTimer>
#include <QJsonDocument>
#include <QJsonObject>
#include <QHostInfo>

#include "base/ConfigDocumentConverter.cpp"
#include "base/ConfigSelector.h"
#include "base/ConfigMetaConverter.h"
#include "base/CoreTypeConverter.h"
#include "base/RcRunId.h"
#include "base/RunDocument.h"

#include "JsonDB.h"
#include "MongoConsts.h"
#include "PostgresDB.h"
#include "RootConfigUtil.h"

#ifdef MONGO_DRIVERS
#include "MongoDB.h"
#include "mongoqt/BsonQtConverter.h"
#endif

BaseConfig::BaseConfig()
{
    CoreTypeConverter::registerTypes();
    AbstractConfigConverter<ConfigDocument>::registerTypes();

    dbList << QSharedPointer<AbstractDB>(new PostgresDB);
#ifdef MONGO_DRIVERS
    dbList << QSharedPointer<AbstractDB>(new MongoDB);
#endif
    dbList << QSharedPointer<AbstractDB>(new JsonDB);
}

void BaseConfig::connectToDb()
{
    for(auto &db : dbList)
        db->connectToDb();
}

ConfigDocument BaseConfig::readConfigDocument(const ConfigSelector &cs)
{
    if(cs.getType().isEmpty()){
        qCritical() << "Can't read config with empty program type";
        return {};
    }

    for(auto &db : dbList) {
        if(db->isAlive()) {
            auto doc = db->readConfigDocument(cs);
            if(!doc.isEmpty())
                return doc;
        }
    }

    return {};
}

QJsonObject BaseConfig::readConfig(const ConfigSelector &cs)
{
    auto cd = readConfigDocument(cs);
    if(cd.getConfigMeta().isPhantom())
        return {};
    return cd.getConfig();
}

bool BaseConfig::writeConfigDocument(ConfigDocument &cd)
{
    if (cd.isEmpty()) {
        qCritical() << "Can't write config with empty program type";
        return false;
    }

    bool ok = true;
    for(auto &db : dbList) {
        if(db->isAlive()){
            ok &= db->writeConfigDocument(cd);
        }
    }
    return ok;
}

bool BaseConfig::writeConfig(const ConfigSelector &cs, const QJsonObject &jsonCfg)
{
    if (cs.getType().isEmpty()) {
        qCritical() << "Can't write config with empty program type";
        return false;
    }

    ConfigDocument cd(cs);
    cd.setConfig(jsonCfg);
    return writeConfigDocument(cd);
}

QMap<ProgramIndex, QMap<ConfigurationName, QDateTime> > BaseConfig::get_config_map(const DaqModule &program_type)
{
    if(program_type.isEmpty()) {
        return {};
    }

    auto mergeResults = [](const QMap<ProgramIndex, QMap<ConfigurationName, QDateTime> > &v1,
                        const QMap<ProgramIndex, QMap<ConfigurationName, QDateTime> > &v2)
    {
        auto ret = v1;
        for(auto prIt=v2.cbegin(); prIt!=v2.cend(); ++prIt) {
            auto &prRet = ret[prIt.key()];
            for(auto cnIt=prIt->cbegin(); cnIt!=prIt->cend(); ++cnIt) {
                auto &d1 = prRet[cnIt.key()];
                if(d1.isNull() || d1<*cnIt)
                    d1 = *cnIt;
            }
        }
        return ret;
    };

    QMap<ProgramIndex, QMap<ConfigurationName, QDateTime> > toReturn;
    for(auto &db :dbList) {
        if(db->isAlive()) {
            auto v = db->get_config_map(program_type);
            toReturn = mergeResults(toReturn, v);
        }
    }

    return toReturn;
}

bool BaseConfig::delete_config(const ConfigSelector &cs)
{
    if(cs.getType().isEmpty()){
        qCritical() << "Can't delete config with empty program type";
        return false;
    }

    auto cm = ConfigMeta();
    cm.setPhantom(true);
    auto cd = ConfigDocument({}, cs, cm);
    bool ok = true;
    for(auto &db : dbList) {
        if(db->isAlive()) {
            ok &= db->writeConfigDocument(cd);
        }
    }

    return ok;
}

bool BaseConfig::writeRunDocument(const RunDocument &runDoc)
{
    bool ok = true;
    auto runDocObj = QVariant::fromValue(runDoc).value<QJsonObject>();
    for(auto db : dbList) {
        if(db->isAlive()) {
            ok &= db->writeConfigRun(runDoc.runId, runDocObj);
        }
    }
    return ok;
}

//------------- LOCAL ----- MONGO ---
// case 1     [normal]    [normal]
// case 2     [normal]    [fantom]
// case 3     [normal]    [      ]
// case 4     [fantom]    [normal]
// case 5     [fantom]    [fantom]
// case 6     [fantom]    [      ]
// case 7     [      ]    [normal]
// case 8     [      ]    [fantom]

#ifdef MONGO_DRIVERS
bool BaseConfig::synchronization(const DaqModule &program_type)
{
    return false;
//    Q_ASSERT(!program_type.isEmpty());
//    qInfo() << "Mongo synchronization to local";
//    if(!local_db.local_db_chekup()){
//        qWarning() << QString("Mongo synchronization failed");
//        return false;
//    }
//    auto list_local = local_db.get_rootconfig_map(program_type);
//    auto list_mongo = mongo_db.get_json_map(program_type);

//    for(auto pi : list_local.keys()) {
//        auto &list_local_pi = list_local[pi];
//        for(auto cn : list_local_pi.keys()) {
//            const auto cs = ConfigSelector(program_type, pi, cn);
//            auto &rc_local = list_local_pi[cn];
//            QDateTime local_date = RootConfigUtil::getDeletionTimestamp(rc_local);
//            bool local_dead = !local_date.isNull();
//            if(!local_dead)
//                local_date = RootConfigUtil::getModificationTimestamp(rc_local);

//            bool contains = list_mongo.contains(pi) && list_mongo[pi].contains(cn);
//            if(!contains) {
//                if(local_dead){
//                    local_db.delete_config(cs); // case 6
//                }
//                else{
//                    mongo_db.write_config(rc_local.get_json().object()); // case 3
//                }
//                continue;
//            }

//            auto &rc_mongo = list_mongo[pi][cn];
//            QDateTime mongo_date = RootConfigUtil::getDeletionTimestamp(rc_mongo);
//            bool mongo_dead = !mongo_date.isNull();
//            if(!mongo_dead)
//                mongo_date = RootConfigUtil::getModificationTimestamp(rc_mongo);
//            bool local_is_newer = local_date > mongo_date;
//            if(local_is_newer) {
//                if(local_dead) {
//                    mongo_db.delete_config(cs); // case 4 and 5
//                    local_db.delete_config(cs);
//                }
//                else{
//                    mongo_db.write_config(rc_local.get_json().object()); // case 1 and 2
//                }

//            } else {
//                if(mongo_dead) {
//                    local_db.delete_config(cs); // case 2 and 5
//                    list_local_pi.remove(cn);
//                }
//                else{
//                    if(rc_local != BsonQtConverter::to_rootconfig(rc_mongo))
//                        local_db.write_config(BsonQtConverter::to_rootconfig(rc_mongo), false); // case 1 and 4
//                }
//            }
//        }
//    }
//    local_db.sync();

//    for(auto pi : list_mongo.keys()) {
//        auto &list_mongo_pi = list_mongo[pi];
//        for(auto cn : list_mongo_pi.keys()) {
//            auto &rc_mongo = list_mongo_pi[cn];
//            bool contains = list_local.contains(pi) && list_local[pi].contains(cn);
//            if(contains) continue; // case 1, 2, 4, 5
//            if(RootConfigUtil::is_phantom(rc_mongo)) continue; // case 8
//            local_db.write_config(BsonQtConverter::to_rootconfig(rc_mongo)); // case 7
//        }
//    }
//    return true;
}
#endif
