/*
**    Copyright 2011-2022 Ivan Filippov
**    Copyright 2011-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "MlinkStreamReceiver.h"

#include <stdexcept>
#include <iostream>
#include <cstdio>
#include <cstring>
#include <cerrno>
#include <unistd.h>

#include <QCoreApplication>
#include <QDebug>
#include <QHostAddress>
#include <QLoggingCategory>
#include <QTime>
#include <QTimer>

#include "RawCheck.h"
#include "base/HwipSocketAddress.h"
#include "PacketRingBuffer.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc_ms_rx)
Q_LOGGING_CATEGORY(lc_ms_rx, "afi.mstream.rx")
}

using std::cerr;
using std::endl;
namespace {
const int RUN_LOOP_PERIOD_MS = 1000;
const int SEND_WORK_TIME_MS = 10;
const int WORK_SIZE_LIMIT = 100;
const int RECV_SELECT_TIMEOUT_US = 2e3; // 1 ms
const int RUN_TIME_CHECK_PERIOD_US = RUN_LOOP_PERIOD_MS*1e3/100; // Check run time 10 times
const int RUN_TIME_CHECK_PERIOD_CNT = RUN_TIME_CHECK_PERIOD_US/RECV_SELECT_TIMEOUT_US;
}

MlinkStreamReceiver::MlinkStreamReceiver(DeviceIndex index, QSharedPointer<PacketRingBuffer> ringBuf,
                                         bool _bigFragId, quint16 hwBufSize, bool _doPrintStat)
    : QObject(nullptr),
      sharedBuffer(ringBuf),
      deviceIndex(index),
      mstream(index.getIdent(), RECV_SELECT_TIMEOUT_US/2),
      bigFragId(_bigFragId),
      runLoopTime(new QElapsedTimer()),
      runLoopTimer(new QTimer(this)),
      rc(new RawCheck),
      sendWorkTimer(new QTimer(this)),
      RECV_LIMIT(hwBufSize*4),
      msgsTt(RECV_LIMIT),
      iovecsTt(RECV_LIMIT),
      stat(index, RECV_LIMIT),
      timeoutCnt(0),
      runTimeCheckPeriodCnt(RUN_TIME_CHECK_PERIOD_CNT),
      doPrintStat(_doPrintStat)
{
    setObjectName(QString("MR_%1").arg(deviceIndex.getSerialStr()));

    connect(runLoopTimer, &QTimer::timeout, this, &MlinkStreamReceiver::runLoop);
    runLoopTimer->setSingleShot(true);
    runLoopTimer->start(0);
    if(bigFragId)
        rc->setBigFragId();
#if (! SEM_INTERACT)
    connect(sendWorkTimer, &QTimer::timeout, this, &MlinkStreamReceiver::sendWork);
    sendWorkTimer->setSingleShot(true);
    sendWorkTimer->setInterval(SEND_WORK_TIME_MS);
#endif
    QTimer::singleShot(0, this, &MlinkStreamReceiver::deferredInit);
    qDebug(lc_ms_rx).noquote() << QString("[%1] receiver created").arg(deviceIndex.getIdent());
}

MlinkStreamReceiver::~MlinkStreamReceiver()
{
    setDeleteFlag();
    disconnectFromHardware();
    runLoopTimer->stop();

    QElapsedTimer timer;
    int loop=0;
    while (sharedBuffer->getTaskToErase() && timer.elapsed()<1000) {
        ++loop;
        QCoreApplication::processEvents(QEventLoop::AllEvents, 100);
//        releaseData()
    }

    if(sharedBuffer->getTaskToErase()){
        qWarning(lc_ms_rx) << "failed to terminate";
    }

    delete runLoopTime;

    qDebug(lc_ms_rx).noquote() << QString("[%1] receiver destroyed").arg(deviceIndex.getIdent());
}

void MlinkStreamReceiver::deferredInit()
{
    qDebug(lc_ms_rx).noquote() << QString("[%1] receiver started").arg(deviceIndex.getIdent());
}

void MlinkStreamReceiver::setDeleteFlag()
{
    deleteFlag.store(true, std::memory_order_relaxed);
}

bool MlinkStreamReceiver::hasDeleteFlag() const
{
    return deleteFlag.load(std::memory_order_relaxed);
}

void MlinkStreamReceiver::hostDiscovered(const QHostAddress &address)
{
    if (mstream.getPeerAddress().hostAddress() == address)
        return;
    disconnectFromHardware();
    mstream.setPeerAddress(address);
    connectToHardware();

    if (!mstream.isConnected())
        connectToHardware();
}

void MlinkStreamReceiver::connectToHardware(bool force)
{
    qDebug(lc_ms_rx) << __PRETTY_FUNCTION__;
    if(force)
        disconnectFromHardware();

    mstream.connectToHardware();
    if (mstream.isBound())
        emit rxPortUpdated(mstream.getLocalAddress().port);
    if (mstream.isConnected())
        emit peerConnected();
}

void MlinkStreamReceiver::disconnectFromHardware()
{
    qDebug(lc_ms_rx) << __PRETTY_FUNCTION__;
    if (!mstream.isConnected())
        return;
    mstream.disconnectFromHardware();
    emit peerDisconnected();
}

void MlinkStreamReceiver::printLoopStat(int loopCount, int loopTime_ms)
{
    int timePerLoopUs = loopTime_ms*1000/loopCount;
    if(doPrintStat) {
        QStringList sizeStat;
        for(int i=0; i<stat.pckSizeCnt.size(); ++i) {
            const quint32 &val = stat.pckSizeCnt[i];
            if(val)
                sizeStat << QString("pckSize%1=%2").arg(i).arg(val);
        }
        std::cerr << QString("[RecvStat]noSpaceLoop=%1;processDidntHelp=%2;"
                             "loopOk=%3;checkFailed=%4;noData=%5;"
                             "LoopCount=%6;LoopTimeMs=%7;Time_us/Loop=%8;%9\n")
                     .arg(stat.noFree).arg(stat.didntHelp)
                     .arg(stat.free).arg(stat.checkFail).arg(stat.noData)
                     .arg(loopCount).arg(loopTime_ms).arg(timePerLoopUs)
                     .arg(sizeStat.join(";"))
                     .toStdString().c_str();
    }
    stat.acquiredAtMs = QDateTime::currentMSecsSinceEpoch();
    emit stat_from_receiver_updated(stat);
    runTimeCheckPeriodCnt = timePerLoopUs ? qMax(1, RUN_TIME_CHECK_PERIOD_US/timePerLoopUs) : 1;
}

ssize_t MlinkStreamReceiver::recvfrom()
{
    mmsghdr *msgsPtr = &msgsTt[0];
    memset(msgsPtr, 0, sizeof(mmsghdr)*RECV_LIMIT);
    uint taskCandidat = sharedBuffer->taskIndexWrite;
    for (int i = 0; i < RECV_LIMIT; i++) {
        iovecsTt[i].iov_base         = sharedBuffer->tasks[taskCandidat].buf;
        iovecsTt[i].iov_len          = BUFLEN;
        msgsTt[i].msg_hdr.msg_iov    = &iovecsTt[i];
        msgsTt[i].msg_hdr.msg_iovlen = 1;
        sharedBuffer->shiftIndex(taskCandidat);
    }

    int retval = mstream.recvmmsg(msgsPtr, RECV_LIMIT);

    if(retval > 0) {
        ++stat.pckSizeCnt[retval];
        taskCandidat = sharedBuffer->taskIndexWrite;
        for(int i=0; i<retval; ++i) {
            sharedBuffer->tasks[taskCandidat].len = msgsTt[i].msg_len;
            sharedBuffer->shiftIndex(taskCandidat);
        }
    } else {
        if(retval==0) {
            if (retval >= 0 && retval < stat.pckSizeCnt.size())
                ++stat.pckSizeCnt[retval];
        } else {
            if(errno == EAGAIN) {
                ++stat.pckSizeCnt[0];
                retval = 0;
            }
        }
    }
    return retval;
}

void MlinkStreamReceiver::runLoop()
{
    runLoopTime->restart();
    stat.clear(static_cast<int>(RECV_LIMIT));
    int loopCount = 0; // reduce number of timer checks, see SELECT_TIMEOUT_US

    while(true) {
        if((++loopCount) % runTimeCheckPeriodCnt == 0){
            int ms = runLoopTime->elapsed();
             if(ms > RUN_LOOP_PERIOD_MS) {
                 printLoopStat(loopCount, runLoopTime->restart());
                 loopCount=0;
             }
             QCoreApplication::processEvents(QEventLoop::ExcludeSocketNotifiers, 1);
             if (hasDeleteFlag()) return;
        }
        if (!mstream.isBound()) {
            runLoopTimer->start(100);
            return;
        }

        // Let timer to do its work
#if SEM_INTERACT
        if(freeSize.available() < RECV_LIMIT) {
            ++stat.noFree;
            if(freeSize.tryAcquire(RECV_LIMIT, 100)) {
                freeSize.release(RECV_LIMIT);
            } else {
                QCoreApplication::processEvents(QEventLoop::ExcludeSocketNotifiers, 0);
                ++stat.didntHelp;
                continue;
            }
        } else
            ++stat.free;
#else
        if(!sharedBuffer->hasFreeTask(RECV_LIMIT)) {
            ++stat.noFree;
            if(sharedBuffer->getTodoTask())
                sendWork();
            if(sharedBuffer->getTaskToErase()) {
                QCoreApplication::processEvents(QEventLoop::ExcludeSocketNotifiers, 0);
                if (hasDeleteFlag()) return;
            }
            if(!sharedBuffer->hasFreeTask(RECV_LIMIT)) {
                ++stat.didntHelp;
                continue;
            }
        } else {
            ++stat.free;
        }
#endif
        const int pckNum = recvfrom();
        auto curIndex = sharedBuffer->taskIndexWrite;
        if(pckNum > 0) {
            for(int i=0; i<pckNum; ++i) {
                PacketBuffer &pkbuf = sharedBuffer->tasks[curIndex];
                sharedBuffer->shiftIndex(curIndex);
                pkbuf.setMStreamV2Packet(); // remove mlink encapsulation, setup payload ptr
                const auto msHdr = pkbuf.first();
                if (msHdr && msHdr->deviceId == deviceIndex.getDevId()) {
                    const auto ackItem = AckItem(*msHdr, true);
                    ackStack.insert(ackItem);
#if (! SEM_INTERACT)
                    if(sharedBuffer->getTodoTask()==0)
                        sendWorkTimer->start();
#endif
                } else {
                    ++stat.checkFail;
                }
            }
            mstream.send_ack(ackStack);
            ackStack.clear();
            sharedBuffer->shiftIndex(sharedBuffer->taskIndexWrite, pckNum);
#if SEM_INTERACT
            dataSize.release(pckNum);
            freeSize.acquire(pckNum);
#endif
        } else {
            if(pckNum == 0) {
                ++stat.noData;
                continue;
            } else {
                continue;
            }
        }

#if (! SEM_INTERACT)
        if(sharedBuffer->getTodoTask() > WORK_SIZE_LIMIT)
            sendWork();
#endif
    }
}

#if (! SEM_INTERACT)
void MlinkStreamReceiver::releaseData(int shift)
{
    Q_ASSERT(sharedBuffer->getTaskToErase() >= shift);
    sharedBuffer->shiftIndex(sharedBuffer->taskIndexErase, shift);
}

void MlinkStreamReceiver::sendWork()
{
    const int todoTask = sharedBuffer->getTodoTask();
    if (hasDeleteFlag() || todoTask==0)
        return;
    sendtasks.clear();
    for(int i=0; i<todoTask; ++i) {
        const auto &task = sharedBuffer->tasks[sharedBuffer->taskIndexSend];
        sendtasks.push_back(task);
        sharedBuffer->shiftIndex(sharedBuffer->taskIndexSend);
    }
    emit gotData(sendtasks);
//    releaseData(sendtasks);
}
#endif

void MlinkStreamReceiver::deleteReceiver()
{
    setDeleteFlag();
    deleteLater();
}
